module LightTool
  PRINT_OBSERVER_TRIGGERED_INFO = (D5Converter::ENVIRONMENT!="test") ? false : false
end

module D5Light
  def D5Light.update_and_send
    D5LightDataManager.update_and_send
    return
  end

  def D5Light.start_light_sync
    D5LightDataManager.start_light_sync
    return
  end

  def D5Light.stop_light_sync
    D5LightDataManager.stop_light_sync
    return
  end

  def self.initialize(model)
    D5LightDataManager.initialize(model)
    LightTool.init(model)
  end
end

module D5LightDataManager
  class LightIterator
    class Node
      attr_accessor :value, :next_node
      def initialize(value,next_node = nil)
        @value = value
        @next_node = next_node
      end
    end

    def initialize(model)
      @model = model
    end

    def paths_of_inst(inst)
      light_paths = []
      each_light_of_inst(inst){ |inst_path| light_paths << inst_path }
      return light_paths
    end

    def paths
      light_paths = []
      @model.definitions.each do |defi|
        if LightTool.getType(defi)!=nil
          defi.instances.each { |inst| light_paths += paths_of_inst(inst) }
        end
      end

      light_paths
    end

    # each_light_of_inst(inst) { |inst_path| ... }
    def each_light_of_inst(inst)
      # 如果灯光不在group或其他组件instance中？
      if inst.parent==Sketchup.active_model
        inst_path = Sketchup::InstancePath.new([inst])
        yield(inst_path) if block_given?
        return
      end

      pathNodeList=[[Node.new(inst,nil)]]
      headList = Array.new()
      while true
        end_search = true
        tmpList = Array.new
        for node in pathNodeList[-1]
          val = node.value

          if val.parent==Sketchup.active_model or val.parent==nil
            headList<<Node.new(0,node) #头节点前的空节点
          else #val.parent==Sketchup::ComponentDefinition
            for par in val.parent.instances
              tmpList<<Node.new(par,node)
              end_search = false
            end
          end
        end
        pathNodeList<<tmpList
        if end_search
          break
        end
      end

      # headlist中存放了多个灯光实例的路径的head，现将每一个灯光路径组合为InstancePath
      # 获取每一个灯光的transformation
      resTransArray = Array.new()
      for head in headList
        cur = head.next_node
        tmpInstPath_array = Array.new
        while cur!=nil
          tmpInstPath_array<<cur.value
          cur = cur.next_node
        end
        inst_path = Sketchup::InstancePath.new(tmpInstPath_array)
        yield(inst_path) if block_given?
      end
    end

    # each_light{ |inst_path, inst, type| ... }
    def each_light
      @model.definitions.each do |defi|
        light_type = LightTool.getType(defi)
        if light_type != nil
          defi.instances.each do |inst|
            each_light_of_inst(inst) { |inst_path| yield(inst_path, inst, light_type) if block_given? }
          end
        end
      end
    end
  end

  class ModelObserver < Sketchup::ModelObserver
    def onPreSaveModel(model)
      D5LightDataManager.save_light_data
    end

    def onTransactionCommit(model)
      D5LightDataManager.update_and_send
    end

    def onTransactionRedo(model)
      D5LightDataManager.update_and_send
    end

    def onTransactionUndo(model)
      D5LightDataManager.update_and_send
    end

    @@model_observer = ModelObserver.new
    def self.model_observer
      @@model_observer
    end
  end

  class LightData
    ATTRI_DIC_NAME = 'pid_path_key_map' # 持久化：记录灯光key数据的词典名
    Y_OFFSET_TRANS = Geom::Transformation.translation(Geom::Vector3d.new(0,0,-1.mm))

    @path_key_map = Hash.new
    def path_key_map
      @path_key_map
    end

    #Plan A1
    # 总是记录着最新的trans，与当前灯光一一对应
    @path_trans_map = Hash.new
    def path_trans_map
      @path_trans_map
    end
    @path_state_map = Hash.new
    def path_state_map
      @path_state_map
    end

    def deal_trans_and_path_change
      new_path_trans_map = Hash.new
      added_path_list = Array.new

      # 遍历找到新增了哪些灯光。
      # 且对已有的灯光更新transformation缓存。
      @light_it.each_light do |inst_path, inst, type|
        inst_id_path = D5LightDataManager.path_to_id_str(inst_path)

        # 将灯向灯光方向偏移一点，避免与模型重面问题
        trans = inst_path.transformation
        trans = trans * Y_OFFSET_TRANS if type != LightTool::POINT_TYPE

        # 下面将transformation转为array是为了便于判断相等；
        # 又因为需要考虑浮点数误差，所以用collect和to_l方法将array中的所有数转为su的length，length对象的比较考虑了误差
        new_trans = trans.to_a.collect{ |item| item.to_l }
        old_trans = @path_trans_map[inst_id_path]

        new_path_trans_map[inst_id_path] = new_trans

        if old_trans
          # 比较修改前后灯的trans, 如果不一样就记下来，表示需要在同步时更新
          if old_trans.to_a != new_trans.to_a
            @path_state_map[inst_id_path] = 1
          end
        else # old_trans == nil
          # old_trans为nil时认为是新增的，并记下来。（新增的包括由于InstancePath变化而新增的，以及实际新增的）
          added_path_list.push(inst_path)
        end

        # 删掉原记录，表示已处理，这样剩下的就是此次修改删除的灯光（对于InstancePath的变化，认为是删除并新增）
        @path_trans_map.delete(inst_id_path) if old_trans
      end

      deleted_map = @path_trans_map # 由于上面循环中的处理，这里剩下的应该就是此次修改删除的灯光。原数据是map，直接当作list来用

      if added_path_list.empty?
        # InstancePath不变, 则可能transformation仿射变换发生了变化
      else
        # 这里假定这样的前提：对于InstancePath变化的灯，其位置应该是不变的；如果位置也变了，则认为是实际新增的
        # 新增的包括由于InstancePath变化而新增的，以及实际新增的
        # 对于前者，其对应的因InstancePath变化而删除的应该在deleted_path_list中，在其中查找与新增的transformation是一致的，认为是变化前后的同一个灯
        # 对于后者，则需要新生成灯光的key
        for added_path in added_path_list
          inst_id_path = D5LightDataManager.path_to_id_str(added_path)

          new_trans = new_path_trans_map[inst_id_path]
          old_id_path = deleted_map.key(new_trans) # 由trans找对应的原来的灯
          # D5Message.d5_puts("new_inst_id_path already has value", 1) if @path_key_map[inst_id_path]!=nil

          # 根据是否找到对应旧path，决定用旧path的key还是新生成key
          key = old_id_path ? @path_key_map[old_id_path] : added_path.to_a.collect!{ |entity| entity.persistent_id }.join('.')+'__'+rand(9999).to_s # persistent_id_path应该不会重复,先加个随机数确保不重复

          # 更新key_map
          @path_key_map[inst_id_path] = key

          # 更新待发送状态
          @path_state_map[inst_id_path] = 1
        end
      end

      @path_trans_map = new_path_trans_map
    end

    def load
      @path_key_map = Hash.new
      @path_trans_map = Hash.new
      @path_state_map = Hash.new

      # 从attribute恢复key_map。attribute中记录的是pid_path-key，需要转为id_path-key
      restored_from_attribute = false
      create_if_empty = false
      dictionary = @model.attribute_dictionary(ATTRI_DIC_NAME, create_if_empty)
      if dictionary
        for pid_path,key in dictionary
          begin
            inst_path = @model.instance_path_from_pid_path(pid_path)
            inst_id_path = D5LightDataManager.path_to_id_str(inst_path)
            @path_key_map[inst_id_path] = key
          rescue
            # 跳过
            D5Message.d5_puts("invalid pid path in diction: #{pid_path}", 1)
          end
        end

        restored_from_attribute = true
        D5Message.d5_puts("light data restored from attribute.")
      end

      @model.definitions.each do |defi|
        type = LightTool.getType(defi)
        if type!=nil
          defi.instances.each do |inst|
            paths = @light_it.paths_of_inst(inst)
            for i in 0...paths.size
              inst_path = paths[i]
              inst_id_path = D5LightDataManager.path_to_id_str(inst_path)
              # 兼容旧版本的灯光数据：按照旧规则生成key
              if @path_key_map[inst_id_path].nil?
                # 如果前面从文件恢复了，那么这里不应出现找不到灯光key的情况
                D5Message.d5_puts("#{inst_id_path} can't find the key. A new one is made.", 1) if restored_from_attribute

                light_id = inst.persistent_id.to_s + "_#{i}" # inst = inst_path.to_a[-1]
                @path_key_map[inst_id_path] = light_id
              end

              # 将灯向灯光方向偏移一点，避免与模型重面问题
              trans = inst_path.transformation
              trans = trans * Y_OFFSET_TRANS if type != LightTool::POINT_TYPE

              # Plan A1: 缓存所有灯的位置 Hash{InstancePath, Trans}
              # 下面将transformation转为array是为了便于判断相等；
              # 又因为需要考虑浮点数误差，所以用collect和to_l方法将array中的所有数转为su的length，length对象的比较考虑了误差
              @path_trans_map[inst_id_path] = trans.to_a.collect{ |item| item.to_l }

              # 缓存所有灯的数据状态
              @path_state_map[inst_id_path] = 1 # state_add
            end
          end
        end
      end
    end

    def save
      # 灯光key持久化：保存文件前，写入attribute
      @model.start_operation('', true, false, true)
      @model.attribute_dictionaries.delete(ATTRI_DIC_NAME) # 先删除旧的，再新建一个
      create_if_empty = true
      dictionary = @model.attribute_dictionary(ATTRI_DIC_NAME, create_if_empty)

      for path in @light_it.paths
        pid_path = path.to_a.collect!{ |entity| entity.persistent_id }.join('.') # path.persistent_id_path得到的不完整，所以改为自己生成
        key = @path_key_map[D5LightDataManager.path_to_id_str(path)]
        dictionary[pid_path] = key
      end
      @model.commit_operation
    end

    def initialize(model)
      @model = model
      @light_it = D5LightDataManager::LightIterator.new(model)
      D5Benchmark::bm("Light Data Load And Init") { load }
    end
  end

  class SyncData
    module XmlHelper
      def definition_xml_name(definition)
        defname = definition ? definition.name : "default"
        # 兼容旧版本数据？
        if defname.length>=14 and defname[0,14]=="D5RenderLight."
          defname = defname[14,defname.length-14]
        end
        return defname
      end

      TO_YXZ_TRANS = Geom::Transformation.axes(ORIGIN, Y_AXIS, X_AXIS, Z_AXIS)
      def getXmlTransStr(trans_array)
        trans = Geom::Transformation.new(trans_array)

        # 处理非正交变换矩阵
        if (trans_array[0] * trans_array[4] + trans_array[1] * trans_array[5] + trans_array[2] * trans_array[6]).abs > 1e-5 \
        || (trans_array[0] * trans_array[8] + trans_array[1] * trans_array[9] + trans_array[2] * trans_array[10]).abs > 1e-5 \
        || (trans_array[4] * trans_array[8] + trans_array[5] * trans_array[9] + trans_array[6] * trans_array[10]).abs > 1e-5
          # 重新计算灯的坐标轴变换
          tar_x = trans.xaxis # x轴不变
          tar_z = trans.xaxis * trans.yaxis # z轴取xy平面的法线方向
          tar_y = tar_x * tar_z # y轴根据新的x和z重新计算
          fixed_trans = Geom::Transformation.axes(trans.origin, tar_x.normalize, tar_y.normalize, tar_z.normalize) # 最后三轴标准化
          # 处理缩放
          scale_trans = Geom::Transformation.scaling(trans.xaxis.length / 1, trans.yaxis.length / 1,  trans.zaxis.length / 1)
          trans = fixed_trans * scale_trans
        end

        # 左右手坐标系转换。相当于
        # swapTrans=[
        #     ot[5],ot[4],ot[6],ot[7],
        #     ot[1],ot[0],ot[2],ot[3],
        #     ot[9],ot[8],ot[10],ot[11],
        #     ot[13],ot[12],ot[14],ot[15]
        # ]
        yxz_trans = TO_YXZ_TRANS * trans * TO_YXZ_TRANS #TO_YXZ_TRANS * Geom::Transformation.new(trans_array) * TO_YXZ_TRANS

        # 数值类型转为float；单位 英寸转厘米
        ot = yxz_trans.to_a.collect{ |item| item.to_f }#Origin Trans
        for i in 12..14
          ot[i]*=2.54
        end

        return ot.to_s[1...-1]
      end

      MM2INCH = 0.0393701
      #return sizex,sizey,brightness
      def getLightParam(deftype, trans_array)
        if deftype==LightTool::POINT_TYPE
          return nil,nil,"45.0"
        elsif deftype==LightTool::SPOT_TYPE
          return nil,nil,"235.2"
        elsif deftype==LightTool::STRIP_TYPE
          trans = Geom::Transformation.new(trans_array)
          edgeX = Geom::Point3d.new(-MM2INCH*750.0,0,0)
          edgeY = Geom::Point3d.new(0,-MM2INCH*15.0,0)
          edgeX.transform! trans
          edgeY.transform! trans
          new_xaxis = trans.origin-edgeX
          new_yaxis = trans.origin-edgeY
          sizeX=(new_xaxis.length.to_cm*2).to_s
          sizeY=(new_yaxis.length.to_cm*2).to_s
          return sizeX,sizeY,"30.0"
        elsif deftype==LightTool::RECT_TYPE
          trans = Geom::Transformation.new(trans_array)
          edgeX = Geom::Point3d.new(-MM2INCH*500.0,0,0)
          edgeY = Geom::Point3d.new(0,-MM2INCH*500.0,0)
          edgeX.transform! trans
          edgeY.transform! trans
          new_xaxis = trans.origin-edgeX
          new_yaxis = trans.origin-edgeY
          sizeX=(new_xaxis.length.to_cm*2).to_s
          sizeY=(new_yaxis.length.to_cm*2).to_s
          return sizeX,sizeY,"30.0"
        end
      end

      def getLightXML(light_id, parent_group_id, inst, trans, type)
        #light_name = format("%s [%s]",definition_xml_name(inst.definition), inst.persistent_id.to_s)
        light_name = format("%s [%s]", inst.name, definition_xml_name(inst.definition))
        trans_str = getXmlTransStr(trans)
        size_x, size_y, brightness = getLightParam(type, trans)

        light_format = "<LightSource>"
        light_format << "<UpdateMode>1</UpdateMode>"
        light_format << "<LightId>" << light_id << "</LightId>"
        light_format << "<ParentGroupId>" << parent_group_id << "</ParentGroupId>"
        light_format << "<LightName>" << light_name << "</LightName>"
        light_format << "<Transform>" << trans_str << "</Transform>"
        light_format << "<type>" << type.to_s << "</type>"
        light_format << "<SizeX>" << size_y.to_s << "</SizeX>" # 这里x,y互换了，因为SU和D5中坐标系不同。
        light_format << "<SizeY>" << size_x.to_s << "</SizeY>"
        light_format << "<Brightness>" << brightness << "</Brightness>"
        light_format << "</LightSource>"
      end
    end
    include XmlHelper

    class DefiXmlData
      def initialize
        @light_xml_map = Hash.new
        @group_xml = ""
        @group_id = ""
      end
      attr_accessor :light_xml_map
      attr_accessor :group_xml
      attr_accessor :group_id
    end

    def initialize(model, light_data)
      @model = model
      @light_data = light_data
      @light_it = D5LightDataManager::LightIterator.new(model)
      @xml_cache_struct = Hash.new
      @last_send_xml = nil
    end

    def update_xml_to_cache
      # 创建新的xml数据结构
      new_xml = Hash.new

      # 遍历灯光inst_path，为每一项生成instsXml。
      #   - 对于不需要更新xml的灯，从缓存中获得xml。
      #   - 对于需要更新xml的灯，重新生成xml。
      # 需要更新的情况：灯所属组件的引用实例数变化 1->n or n->1 ;灯光的状态标记为需要更改的(trans变化，或Instancepath变化)
      @model.definitions.each do |defi|
        light_type = LightTool.getType(defi)
        if light_type != nil
          defi_id = defi.entityID
          new_xml[defi_id] = DefiXmlData.new
          @xml_cache_struct[defi_id] = DefiXmlData.new if @xml_cache_struct[defi_id] == nil

          new_defi_data = new_xml[defi_id]
          old_defi_data = @xml_cache_struct[defi_id]

          # 处理灯xml
          new_lightXml_map = new_defi_data.light_xml_map # 创建新的instxml_map，储存每一个inst的xml
          old_lightXml_map = old_defi_data.light_xml_map # 旧的instxml_map，用于提高效率
          defi.instances.each do |inst|
            # 处理每一个灯的xml
            @light_it.each_light_of_inst(inst) do |inst_path|
              inst_id_path = D5LightDataManager.path_to_id_str(inst_path)
              need_update = (@light_data.path_state_map[inst_id_path] != 0)
              if need_update || old_lightXml_map.empty?
                # 灯光变化，更新xml
                light_id = @light_data.path_key_map[inst_id_path]
                trans = @light_data.path_trans_map[inst_id_path]
                new_lightXml_map[inst_id_path] = getLightXML(light_id,"%s", inst, trans, light_type) # parent_group_id 先用占位符，下面再填上
              else
                # 灯光没变，从缓存读取xml
                if old_lightXml_map[inst_id_path]
                  new_lightXml_map[inst_id_path] = old_lightXml_map[inst_id_path]
                else
                  # 不应出现的情况
                  D5Message.d5_puts("Light xml is not found in cache! Make a new one.", 2)
                  light_id = @light_data.path_key_map[inst_id_path]
                  trans = @light_data.path_trans_map[inst_id_path]
                  new_lightXml_map[inst_id_path] = getLightXML(light_id,"%s", inst, trans, light_type)
                end
              end
            end
          end

          # 对比修改前后，组件中灯的数量。判断group是否需要修改
          new_count = new_lightXml_map.count
          old_count = old_lightXml_map.count
          if new_count <= 1 && old_count > 1
            group_need_change = true
          elsif new_count > 1 && old_count <= 1
            group_need_change = true
          else
            group_need_change = false
          end

          # 处理group_id和组xml。以便在下一步打包xml时使用
          if group_need_change || old_lightXml_map.empty?
            if new_count > 1
              group_id_str = defi_id.to_s
              defname = definition_xml_name(defi)
              group_xml = "<LightGroup><GroupId>#{group_id_str}</GroupId><ParentGroupId>0</ParentGroupId><GroupName>#{defname}</GroupName></LightGroup>"
            else
              group_id_str = "0" # 0 is root group id
              group_xml = ""
            end
          else
            group_id_str = old_defi_data.group_id
            group_xml = old_defi_data.group_xml
          end
          new_defi_data.group_id = group_id_str
          new_defi_data.group_xml = group_xml
        end
      end

      # 更新并返回 @xml
      @xml_cache_struct = new_xml
    end

    def package_xml
      light_xml = ""
      group_xml = ""
      @xml_cache_struct.each do |defi_id, defi_xml_data|
        lightXml_map = defi_xml_data.light_xml_map
        group_id = defi_xml_data.group_id
        lightXml_map.each do |inst_id_path, instXml|
          light_xml << format(instXml, group_id) if instXml!=nil
        end

        group_xml << defi_xml_data.group_xml
      end

      if light_xml.empty?
        all_xml = ""
      else
        all_xml = "<Root>"
        all_xml<<"<LightGroup><GroupId>0</GroupId><ParentGroupId></ParentGroupId><GroupName>#{@model.title}.skp</GroupName></LightGroup>" # root group
        all_xml<<group_xml
        all_xml<<light_xml
        all_xml<<"</Root>"
      end

      return all_xml
    end

    def update_and_send_xml
      if $d5Converter_connectionStatus==false
        return
      end

      # light entire data to xml data
      update_xml_to_cache
      # xml data to string
      all_xml = package_xml

      # send string
      if all_xml != @last_send_xml
        D5dllFunc::SendLightsDataXml.call($d5Converter_modelPtr, Sketchup.active_model.path.encode("utf-16le"),all_xml.encode("utf-16le"))
        @last_send_xml = all_xml
      end
    end
  end

  def self.path_to_id_str(path)
    id_str = ""
    path.to_a.each { |entity| id_str << entity.entityID.to_s << '.' }
    id_str.chop!
  end

  # 延迟调用
  @@delay_excuter = D5DelayExcuter.new

  def self.update_and_send
    # 延迟更新和发送，以处理用户一次操作中多次触发更新的情况。
    @@delay_excuter.excute do
      D5Benchmark::bm("Light Update") { @@light_data.deal_trans_and_path_change }
      D5Benchmark::bm("Light Sync") { @@sync_data.update_and_send_xml } if @@sync_data
    end
  end

  def self.start_light_sync
    # [暂时删掉，考虑到这个name是否有必要]检查修改所有灯光的name为persistent_id。为了将D5中显示的名称与su中的灯光实例name对应，在这里修改name。
    # @model.start_operation('', true, false, true) # "光源同步初始化" #D5Localize.info("LIGHTTIP_SYNC"]
    # @model.definitions.each { |definition| definition.instances.each { |inst| inst.name = inst.persistent_id.to_s if inst.name!=inst.persistent_id.to_s } if LightTool.getType(definition)!=nil }
    # @model.commit_operation

    @@sync_data = SyncData.new(@model, @@light_data)

    # 更新并发送数据
    self.update_and_send
  end

  def self.stop_light_sync
    @@sync_data = nil
  end

  def self.save_light_data
    D5Benchmark::bm("Light Data Save") { @@light_data.save }
  end

  def self.model
    @model
  end

  def self.initialize(model)
    @model = model
    @@light_data = LightData.new(model)
    @@sync_data = nil

    model.add_observer ModelObserver::model_observer
  end

end

$d5currentInst = nil
$d5doubleClick = false
$d5blue = Sketchup::Color.new(0, 128, 255)
$d5green = Sketchup::Color.new(20,200,100)
$d5transparent = Sketchup::Color.new(0,128,255)
$d5transparent.alpha = 75
$d5pixel = 12

module LightTool
  class InstObserver < Sketchup::InstanceObserver
    @@inst_observer = InstObserver.new
    def self.inst_observer
      @@inst_observer
    end

    def onOpen(instance)
      # 开始编辑组件
      # Sketchup.active_model.start_operation(D5Localize.info("LIGHT_ACTIVATE"],true)
      # $d5currentInst=instance
      if Sketchup.active_model.active_path==nil
        return
      end

      # 判断是否为灯光
      deftype = LightTool.getType(Sketchup.active_model.active_path[-1].definition)
      if deftype==nil
        return
      end

      # 记录状态，在灯光Tool的操作时用到
      $d5doubleClick = true
      $d5currentInst = Sketchup.active_model.active_path[-1]

      # 退出组件编辑状态
      Sketchup.active_model.start_operation('', true, false, true) # "光源同步初始化" #D5Localize.info("LIGHTTIP_SYNC"]
      Sketchup.active_model.close_active
      Sketchup.active_model.commit_operation

      # 激活灯光Tool，进入灯光（位置、方向）修改状态
      # D5RenderLight.Point/ ....Rect/ ....Strip / ....Spot
      if deftype==LightTool::POINT_TYPE
        Sketchup.active_model.select_tool($d5pointTool)
      elsif deftype==LightTool::SPOT_TYPE
        Sketchup.active_model.select_tool($d5spotTool)
      elsif deftype==LightTool::STRIP_TYPE
        Sketchup.active_model.select_tool($d5stripTool)
      elsif deftype==LightTool::RECT_TYPE
        Sketchup.active_model.select_tool($d5rectTool)
      end
    end

    def onClose(instance)
    end
  end

  class DefObserver <Sketchup::DefinitionObserver
    @@definition_observer = DefObserver.new
    def self.definition_observer
      @@definition_observer
    end

    # 为了处理撤销删除的情况，需要用onChangeEntity。又因为onChangeEntity也可以处理新增实例的情况，故不需要onComponentInstanceAdded。
    def onChangeEntity(definition)
      puts "onChangeDefinitionEntity(#{definition})" if PRINT_OBSERVER_TRIGGERED_INFO
      if definition.deleted?
        return
      end

      if LightTool.getType(definition)!=nil
        definition.instances.each do |inst|
          inst.add_observer(InstObserver::inst_observer)
        end
      end
    end
  end

  class DefsObserver < Sketchup::DefinitionsObserver
    @@defs_observer = DefsObserver.new
    def self.defs_observer
      @@defs_observer
    end

    def onComponentAdded(definitions, definition)
      puts "onComponentAdded: #{definition}" if PRINT_OBSERVER_TRIGGERED_INFO

      if definition.deleted?
        # 复制并粘贴一个后，再粘贴时是创建新的组件，此时若在鼠标选择粘贴位置状态时退出，会进这里
        return
      end

      if LightTool.getType(definition)!=nil
        definition.instances.each do |instance|
          instance.add_observer(InstObserver::inst_observer)
        end
        definition.add_observer(DefObserver::definition_observer)
      end
    end
  end

  # 添加Observer。
  def self.init(model)
    defs = model.definitions
    defs.add_observer(DefsObserver::defs_observer)

    for definition in defs
      defname = definition.name
      if defname[0,19].eql?("D5RenderLight.Point")
        LightTool.setType definition,LightTool::POINT_TYPE
      elsif defname[0,18].eql?("D5RenderLight.Spot")
        LightTool.setType definition,LightTool::SPOT_TYPE
      elsif defname[0,19].eql?("D5RenderLight.Strip")
        LightTool.setType definition,LightTool::STRIP_TYPE
      elsif defname[0,18].eql?("D5RenderLight.Rect")
        LightTool.setType definition,LightTool::RECT_TYPE
      end

      # 添加定义observe，和实例observer
      if LightTool.getType(definition)!=nil
        for inst in definition.instances
          inst.add_observer(InstObserver::inst_observer)
        end
        definition.add_observer(DefObserver::definition_observer)
      else
        # 不是灯光的组件定义或群组定义，不需要添加observer
      end
    end
  end

  NONE = 0 #没有待编辑的instance，准备新建
  CREATING = 1 #正在创建中
  STATIC = 2 #准备开始编辑
  MOVING = 3 #正在移动中
  EDITING = 4 #正在编辑中
  RESIZING = 5 #正在更改大小

  POINT_TYPE = 0
  SPOT_TYPE = 1
  STRIP_TYPE = 2
  RECT_TYPE = 3

  def LightTool.addInst(compDef,originTrans) #add instance to current selection when selection is a group or instance
    # sel = Sketchup::active_model.selection
    # if !sel.empty? && sel.single_object?
    #   obj = sel.first
    #   if obj.is_a?(Sketchup::ComponentInstance)
    #     return obj.definition.entities.add_instance compDef,originTrans*obj.transformation.inverse
    #   elsif obj.is_a?(Sketchup::Group)
    #     return obj.entities.add_instance compDef,originTrans*obj.transformation.inverse
    #   end
    # end
    path = Sketchup.active_model.active_path
    inst = nil
    if path==nil
      inst = Sketchup.active_model.entities.add_instance(compDef,originTrans)
    else
      baseTrans = Sketchup::InstancePath.new(path).transformation
      relativeTrans = originTrans*baseTrans.inverse
      inst = path[-1].definition.entities.add_instance(compDef,originTrans)
    end

    # 设置inst的name为id
    #inst.name = inst.persistent_id.to_s # entityID每次打开会变，就需要修改这个name导致即使用户没作更改也会提示保存文件。所以改成persistent_id。

    return inst
  end

  def LightTool.setType(definition, type)
    dict = definition.attribute_dictionary("D5RenderLight",true)
    dict['type']=type
  end
  def LightTool.getType(definition)
    dict = definition.attribute_dictionary("D5RenderLight",false)
    if dict==nil
      return nil
    end
    return dict['type']
  end

  class SpotTool
    attr_reader :direction, :vertex, :mouse_ip
    def getPosition
      return @vertex
    end

    def initialize
      @status = LightTool::NONE
      @vertex = nil
      @direction = nil
      @mouse_ip = Sketchup::InputPoint.new
    end
    def getVertexDirection(inst) #从entites中计算中心点和顶点位置
      if inst==nil or inst.deleted?
        return
      end
      @vertex = nil
      # ents = inst.definition.entities
      # face = ents[72]
      # circleCenter = face.bounds.center
      # puts circleCenter
      mm2inch = 0.0393701
      height = 200*mm2inch
      o = ORIGIN
      regular_center = Geom::Point3d.new(0,0,-height*7.5)
      trans=$d5currentInst.transformation
      @vertex = o.transform(trans)
      @direction = regular_center.transform(trans)
      # circleCenter.transform!($d5currentInst.transformation)
      # @direction = circleCenter+(circleCenter-@vertex)
    end

    def getTrans
      # trans = Geom::Transformation.new(@direction-@vertex)
      # puts trans.to_a

    end


    def activate
      if true==$d5doubleClick
        @status = LightTool::STATIC
        getVertexDirection($d5currentInst)
        $d5doubleClick = false
      else
        $d5currentInst = nil
        @status = LightTool::NONE
      end
    end

    def deactivate(view)
      view.invalidate
    end
    def getExtents
      bb = Sketchup.active_model.bounds
      if @mouse_ip.valid?
        bb.add(@mouse_ip.position)
      end
      if @direction!=nil
        bb.add(@direction)
      end
      if @vertex!=nil
        bb.add(@vertex)
      end
      return bb
    end
    def resume(view)
      # puts "resume: view = #{view}"
      view.invalidate
    end

    def suspend(view)
      # puts "suspend: view = #{view}"
    end

    def onCancel(reason, view)
      # puts "MyTool was canceled for reason ##{reason} in view: #{view}"
      # exit tool
      Sketchup.active_model.select_tool(nil)
    end

    def addSpotCompDef(view)
      mm2inch = 0.0393701
      radius = 100*mm2inch
      height = 200*mm2inch
      spotCompDef = Sketchup.active_model.definitions.add "D5RenderLight.Spot"
      LightTool.setType(spotCompDef, LightTool::SPOT_TYPE)
      scale_mask = 127
      spotCompDef.behavior.no_scale_mask = scale_mask


      spotEnts = spotCompDef.entities
      regular_center = Geom::Point3d.new(0,0,-height)
      path = spotEnts.add_circle(regular_center,Z_AXIS,radius)
      edgePoint = path[0].vertices[0]
      tri = spotEnts.add_face([ORIGIN,regular_center,edgePoint])
      tri.followme(path)
      # offsetCenter = Geom::Point3d.new(offset.x*ratio,offset.y*ratio,offset.z*ratio)
      # path = spotEnts.add_circle(offsetCenter,offset,1)
      # edgePoint = path[0].vertices[0]
      # tri = spotEnts.add_face([ORIGIN,offsetCenter,edgePoint])

      return spotCompDef
    end

    def updateUnderSurface #改变了底面圆心，更新圆锥位置

      new_origin=@vertex
      new_zaxis=@vertex-@direction
      new_trans = Geom::Transformation.new(new_origin,new_zaxis)
      $d5currentInst.transformation = new_trans

      mm2inch = 0.0393701
      height = 200*mm2inch


      regular_direction = Geom::Point3d.new(0,0,-height*7.5)
      @direction = regular_direction.transform(new_trans)

      # spotEnts=$d5currentInst.definition.entities
      # offset=@direction-@vertex
      # ratio = 2.0/offset.length
      # offsetCenter = Geom::Point3d.new(offset.x*ratio,offset.y*ratio,offset.z*ratio)


      # spotEnts.clear!
      # path = spotEnts.add_circle(offsetCenter,offset,1)
      # edgePoint = path[0].vertices[0]
      # tri = spotEnts.add_face([ORIGIN,offsetCenter,edgePoint])
      # tri.followme(path)
      #

      # putKeyPoints($d5currentInst)
    end

    def onLButtonDown(flags, x, y, view)

      case @status
      when LightTool::NONE
        @mouse_ip.pick(view,x,y)
        point=@mouse_ip.position
        Sketchup.active_model.start_operation(D5Localize.info("LIGHTTIP_SPOT"),true)
        @vertex = point
        @status = LightTool::CREATING
      when LightTool::CREATING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@vertex))
        point=@mouse_ip.position
        if @vertex!=point
          @direction = point
          # 添加定义 # TODO: 其他的灯也改成这样，分两步创建定义和实例
          spotCompDef = addSpotCompDef(view)

          #添加实例
          new_origin=@vertex
          new_zaxis=@vertex-@direction
          trans = Geom::Transformation.new(new_origin,new_zaxis)
          mm2inch = 0.0393701
          height = 200*mm2inch
          regular_direction = Geom::Point3d.new(0,0,-height*7.5)
          @direction = regular_direction.transform(trans)
          inst = LightTool.addInst spotCompDef,trans

          $d5currentInst = inst
    
          view.invalidate
          Sketchup.active_model.commit_operation()# 提交更改，以支持撤销重做。

          @status = LightTool::STATIC
        end
      when LightTool::STATIC
        @mouse_ip.pick(view,x,y)
        ph = view.pick_helper
        # These do not require init()
        if ph.test_point(@vertex,x,y,$d5pixel)
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_MOVE"),true)
          @status = LightTool::MOVING
          view.invalidate
        elsif ph.test_point(@direction,x,y,$d5pixel)
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_EDIT"),true)
          @status = LightTool::EDITING
          view.invalidate
        else
          # exit tool
          Sketchup.active_model.select_tool(nil)
        end
      when LightTool::MOVING
        # offset = point-@vertex
        # @direction=@direction+offset
        # new_trans = Geom::Transformation.new(offset)
        # $d5currentInst.transform! new_trans
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        point=@mouse_ip.position
        if @direction!=point
          @vertex = point
          updateUnderSurface
          @status = LightTool::STATIC
          view.invalidate
          Sketchup.active_model.commit_operation
        end
      when LightTool::EDITING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@vertex))
        point=@mouse_ip.position
        if @vertex!=point
          @direction = point
          updateUnderSurface
          @status = LightTool::STATIC
          view.invalidate
          Sketchup.active_model.commit_operation
        end
      end
    end

    def onMouseMove(flags, x, y, view)
      case @status
      when LightTool::NONE
        @mouse_ip.pick(view,x,y)
        view.invalidate
      when LightTool::STATIC
        @mouse_ip.pick(view,x,y)
        view.invalidate
      when LightTool::MOVING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        if @mouse_ip.position!=@direction
          view.invalidate
        end
      when LightTool::EDITING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@vertex))
        if @mouse_ip.position!=@vertex
          view.invalidate
        end
      when LightTool::CREATING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@vertex))
        if @mouse_ip.position!=@vertex
          view.invalidate
        end
      end
    end

    def draw(view)
      view.tooltip=@mouse_ip.tooltip if @mouse_ip.valid?
      getVertexDirection($d5currentInst)
      curmouse=@mouse_ip.position
      mm2inch = 0.0393701
      radius = 100*mm2inch
      height = 200*mm2inch
      pts = Array.new(25)
      for i in 0..23
        arcAngle = i*Math::PI/12
        x = Math.cos(arcAngle)*radius
        y = Math.sin(arcAngle)*radius
        pts[i]=Geom::Point3d.new(x,y,-height)
      end
      pts[24]=Geom::Point3d.new(0,0,0)
      faces = Array.new(24)
      faces[0]=[pts[0],pts[23],pts[24]]
      for i in 1..23
        faces[i]=[pts[i],pts[i-1],pts[24]]
      end

      case @status
      when LightTool::NONE #新建状态下，选择底面中心中
        view.draw_points(@mouse_ip.position, $d5pixel, 2, $d5green)
      when LightTool::CREATING #新建状态下，选择方向标志点中
        view.draw_points(@vertex, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@vertex,curmouse)
        view.line_stipple = "_"
        view.draw_line([@vertex,curmouse])
        view.draw_points(curmouse, $d5pixel, 2, $d5green)
        trans = Geom::Transformation.new(@vertex,@vertex-curmouse)

        for point in pts
          point.transform! trans
        end
        view.drawing_color=$d5transparent
        for face in faces
          view.draw(GL_TRIANGLES,face)
        end
      when LightTool::STATIC #待编辑状态下，底面中心，方向标志点需要画出来
        view.draw_points(@vertex, $d5pixel, 2, $d5blue)
        view.draw_line([@vertex,@direction])
        view.draw_points(@direction, $d5pixel, 2, $d5blue)
      when LightTool::EDITING #编辑状态下，移动direction
        view.draw_points(@vertex, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@vertex,curmouse)
        view.line_stipple = "_"
        view.draw_line([@vertex,curmouse])
        view.draw_points(curmouse, $d5pixel, 2, $d5green)

        trans=$d5currentInst.transformation
        curRotationTrans = Geom::Transformation.new(@vertex,@vertex-@direction)
        backRotationTrans = curRotationTrans.inverse
        newRotationTrans = Geom::Transformation.new(@vertex,@vertex-curmouse)
        for point in pts
          point.transform! trans
          point.transform! backRotationTrans
          point.transform! newRotationTrans
        end
        view.drawing_color=$d5transparent
        for face in faces
          view.draw(GL_TRIANGLES,face)
        end
      when LightTool::MOVING #移动状态下，移动顶点中
        view.draw_points(curmouse, $d5pixel, 2, $d5green)
        view.draw_points(@direction, $d5pixel, 2, $d5blue)
        view.set_color_from_line(curmouse,@direction)
        view.line_stipple = "_"
        view.draw_line([curmouse,@direction])

        trans=$d5currentInst.transformation
        curRotationTrans = Geom::Transformation.new(@vertex,@vertex-@direction)
        backRotationTrans = curRotationTrans.inverse
        newRotationTrans = Geom::Transformation.new(curmouse,curmouse-@direction)
        for point in pts
          point.transform! trans
          point.transform! backRotationTrans
          point.transform! newRotationTrans
        end
        view.drawing_color=$d5transparent
        for face in faces
          view.draw(GL_TRIANGLES,face)
        end
      else
        # puts "error in draw of SpotTool"
      end
    end
  end

  class PointTool
    attr_reader :centerPoint, :mouse_ip
    def getPosition
      return @centerPoint
    end

    def initialize
      @status = LightTool::NONE
      @centerPoint = nil
      @mouse_ip = Sketchup::InputPoint.new
    end

    def getCenterPoint(inst)
      if inst==nil or inst.deleted?
        return
      end
      @centerPoint = inst.bounds.center
    end

    def addPointCompDef(len,view)
      pointCompDef = Sketchup.active_model.definitions.add "D5RenderLight.Point"
      LightTool.setType(pointCompDef, LightTool::POINT_TYPE)
      scale_mask = 127
      pointCompDef.behavior.no_scale_mask = scale_mask
      pointEnts = pointCompDef.entities

      pts = Array.new(6)
      pts[0]=[-len,0,0]
      pts[1]=[0,len,0]
      pts[2]=[len,0,0]
      pts[3]=[0,-len,0]
      pts[4]=[0,0,len]
      pts[5]=[0,0,-len]
      # face = pointEnts.add_face(pts)
      # face.pushpull(-2*len)
      faces = Array.new(8)
      faces[0]=[pts[0],pts[3],pts[4]]
      faces[1]=[pts[0],pts[3],pts[5]]
      for i in 1..3
        faces[2*i]=[pts[i-1],pts[i],pts[4]]
        faces[2*i+1]=[pts[i-1],pts[i],pts[5]]
      end

      for face in faces
        pointEnts.add_face(face)
      end

      inst = LightTool.addInst pointCompDef,Geom::Transformation.new(@centerPoint)
      $d5currentInst = inst

      @status = LightTool::STATIC
      view.invalidate
    end

    def activate
      if true==$d5doubleClick
        #activated by double click
        @status = LightTool::STATIC
        getCenterPoint($d5currentInst)
        $d5doubleClick = false
      else
        $d5currentInst = nil
        @status = LightTool::NONE
      end
    end

    def deactivate(view)
      view.invalidate
    end

    def getExtents
      bb = Sketchup.active_model.bounds
      if @mouse_ip.valid?
        bb.add(@mouse_ip.position)
      end
      if @centerPoint!=nil
        bb.add(@centerPoint)
      end
      return bb
    end

    def onLButtonDown(flags, x, y, view)
      @mouse_ip.pick(view,x,y)
      point=@mouse_ip.position

      # print point.x,point.y,point.z,@mouse_ip.depth,"\n"

      case @status
      when LightTool::NONE
        Sketchup.active_model.start_operation(D5Localize.info("LIGHTTIP_POINT"),true)
        mm2inch = 0.0393701
        side_length = 325*mm2inch
        len = side_length/2
        @centerPoint = point
        addPointCompDef(len,view)
        Sketchup.active_model.commit_operation
      when LightTool::STATIC
        ph = view.pick_helper
        # These do not require init()
        picked = ph.test_point(@centerPoint, x, y, $d5pixel)
        if picked
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_MOVE"),true)
          @status = LightTool::MOVING
          view.invalidate
        else
          # exit tool
          Sketchup.active_model.select_tool(nil)
        end
      when LightTool::MOVING
        offset = point-@centerPoint
        new_trans = Geom::Transformation.new(offset)
        # puts new_trans.to_a
        #update centerPoint and status
        if @mouse_ip.position!=@centerPoint
          $d5currentInst.transform! new_trans
          getCenterPoint($d5currentInst)
          @status = LightTool::STATIC
          Sketchup.active_model.commit_operation
        end
      else
        # puts "wrong in onLButtonDown of point tool"
      end
    end

    def onMouseMove(flags, x, y, view)
      @mouse_ip.pick(view,x,y)

      case @status
      when LightTool::NONE
        view.invalidate
      when LightTool::STATIC
        view.invalidate
      when LightTool::MOVING
        if @mouse_ip.position!=@centerPoint
          view.invalidate
        end
      else
        # puts "wrong in onMouseMove of point tool"
      end

    end

    def draw(view)
      view.tooltip=@mouse_ip.tooltip if @mouse_ip.valid?
      getCenterPoint($d5currentInst)
      curMouse=@mouse_ip.position

      case @status
      when LightTool::NONE
        view.draw_points(curMouse,$d5pixel,2,$d5green)

        mm2inch = 0.0393701
        len = 162.5*mm2inch
        pts = Array.new(6)
        pts[0]=Geom::Point3d.new(-len,0,0)
        pts[1]=Geom::Point3d.new(0,len,0)
        pts[2]=Geom::Point3d.new(len,0,0)
        pts[3]=Geom::Point3d.new(0,-len,0)
        pts[4]=Geom::Point3d.new(0,0,len)
        pts[5]=Geom::Point3d.new(0,0,-len)
        trans = Geom::Transformation.new(curMouse)
        for i in 0..5
          pts[i].transform! trans
        end
        faces = Array.new(8)
        faces[0]=[pts[0],pts[3],pts[4]]
        faces[1]=[pts[0],pts[3],pts[5]]
        for i in 1..3
          faces[2*i]=[pts[i-1],pts[i],pts[4]]
          faces[2*i+1]=[pts[i-1],pts[i],pts[5]]
        end

        view.drawing_color=$d5transparent
        for face in faces
          view.draw(GL_TRIANGLES,face)
        end
      when LightTool::STATIC
        view.draw_points(@centerPoint,$d5pixel,2,$d5blue)
      when LightTool::MOVING
        view.draw_points(curMouse,$d5pixel,2,$d5green)
        view.draw_points(@centerPoint,$d5pixel,2,$d5blue)

        mm2inch = 0.0393701
        len = 162.5*mm2inch
        pts = Array.new
        pts[0]=Geom::Point3d.new(-len,0,0)
        pts[1]=Geom::Point3d.new(0,len,0)
        pts[2]=Geom::Point3d.new(len,0,0)
        pts[3]=Geom::Point3d.new(0,-len,0)
        pts[4]=Geom::Point3d.new(0,0,len)
        pts[5]=Geom::Point3d.new(0,0,-len)
        trans = Geom::Transformation.new(curMouse)
        for i in 0..5
          pts[i].transform! trans
        end
        faces = Array.new(8)
        faces[0]=[pts[0],pts[3],pts[4]]
        faces[1]=[pts[0],pts[3],pts[5]]
        for i in 1..3
          faces[2*i]=[pts[i-1],pts[i],pts[4]]
          faces[2*i+1]=[pts[i-1],pts[i],pts[5]]
        end

        view.drawing_color=$d5transparent
        for face in faces
          view.draw(GL_TRIANGLES,face)
        end
      else
        # puts "wrong in draw of point tool"
      end

    end

    def resume(view)
      # puts "resume: view = #{view}"
      view.invalidate
    end

    def suspend(view)
      # puts "suspend: view = #{view}"
    end

    # Tools can be interrupted for various reasons. In this example tool we
    # simply reset it regardless, but if you need finer granularity you can
    # check the reason code.
    #
    # 0: The user canceled the current operation by hitting the escape key.
    # 1: The user re-selected the same tool from the toolbar or menu.
    # 2: The user did an undo while the tool was active.
    def onCancel(reason, view)
      # puts "MyTool was canceled for reason ##{reason} in view: #{view}"
      # exit tool
      Sketchup.active_model.select_tool(nil)
    end
  end


  # Custom tools create a class which responds to various callback methods from SketchUp.
  class StripTool
    attr_reader :direction, :center, :mouse_ip
    def getDirection(inst)
      if inst==nil or inst.deleted?
        return
      end
      trans=$d5currentInst.transformation
      mm2inch = 0.0393701
      depth = 1500*mm2inch
      o = ORIGIN
      regular_direction = Geom::Point3d.new(0,0,-depth)
      @center = o.transform(trans)
      @direction = regular_direction.transform(trans)
    end

    def initialize
      @status = LightTool::NONE
      @direction = nil
      @center = nil
      @mouse_ip = Sketchup::InputPoint.new
    end


    def activate
      if true==$d5doubleClick
        @status = LightTool::STATIC
        getDirection($d5currentInst)
        $d5doubleClick = false
      else
        $d5currentInst = nil
        @status = LightTool::NONE
      end
    end

    def deactivate(view)
      view.invalidate
    end
    def getExtents
      bb = Sketchup.active_model.bounds
      if @mouse_ip.valid?
        bb.add(@mouse_ip.position)
      end
      if @center!=nil
        bb.add(@center)
      end
      if @direction!=nil
        bb.add(@direction)
      end
      return bb
    end
    def resume(view)
      # puts "resume: view = #{view}"
      view.invalidate
    end

    def suspend(view)
      # puts "suspend: view = #{view}"
    end

    def onCancel(reason, view)
      # puts "MyTool was canceled for reason ##{reason} in view: #{view}"
      # exit tool
      Sketchup.active_model.select_tool(nil)
    end

    def addStripCompDef(view)
      ents = Sketchup.active_model.entities
      stripCompDef = Sketchup.active_model.definitions.add "D5RenderLight.Strip"
      LightTool.setType(stripCompDef, LightTool::STRIP_TYPE)
      stripEnts = stripCompDef.entities

      mm2inch = 0.0393701
      length = 1500*mm2inch
      width = 30*mm2inch
      len = length/2
      wid = width/2
      depth = 1500*mm2inch
      pts=[[-len,-wid,0],[-len,wid,0],[len,wid,0],[len,-wid,0]]
      regular_edgePoints=[Geom::Point3d.new([-len,0,0]),Geom::Point3d.new([0,-wid,0]),Geom::Point3d.new([len,0,0]),Geom::Point3d.new([0,wid,0])]
      regular_direction = Geom::Point3d.new(0,0,-depth)
      stripEnts.add_face(pts)

      stripEnts.add_line(regular_edgePoints[0],regular_edgePoints[2])
      stripEnts.add_line(regular_edgePoints[1],regular_edgePoints[3])

      new_origin=@center
      new_zaxis=@center-@direction
      trans = Geom::Transformation.new(new_origin,new_zaxis)

      inst = LightTool.addInst stripCompDef,trans
      $d5currentInst = inst
      @direction = regular_direction.transform(trans)
      @status = LightTool::STATIC
      view.invalidate
    end

    def updateDirection(new_direction) #更改朝向
      curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
      backRotationTrans = curRotationTrans.inverse
      newRotationTrans = Geom::Transformation.new(@center,@center-new_direction)
      $d5currentInst.transform! backRotationTrans
      $d5currentInst.transform! newRotationTrans

      mm2inch = 0.0393701
      depth = 1500*mm2inch
      regular_direction = Geom::Point3d.new(0,0,-depth)
      @direction = regular_direction.transform($d5currentInst.transformation)
    end

    def updateCenter(new_center)
      curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
      backRotationTrans = curRotationTrans.inverse
      newRotationTrans = Geom::Transformation.new(new_center,new_center-@direction)
      $d5currentInst.transform! backRotationTrans
      $d5currentInst.transform! newRotationTrans

      mm2inch = 0.0393701
      depth = 1500*mm2inch
      regular_direction = Geom::Point3d.new(0,0,-depth)
      @direction = regular_direction.transform($d5currentInst.transformation)
      @center = new_center
    end

    def onLButtonDown(flags, x, y, view)

      case @status
      when LightTool::NONE
        @mouse_ip.pick(view,x,y)
        point=@mouse_ip.position
        Sketchup.active_model.start_operation(D5Localize.info("LIGHTTIP_STRIP"),true)
        @center = point
        @status = LightTool::CREATING
      when LightTool::CREATING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        point=@mouse_ip.position
        if @center!=point
          @direction = point
          addStripCompDef(view)
          @status = LightTool::STATIC
          Sketchup.active_model.commit_operation
        end
      when LightTool::STATIC
        @mouse_ip.pick(view,x,y)
        point=@mouse_ip.position
        ph = view.pick_helper
        # These do not require init()
        if ph.test_point(@center,x,y,$d5pixel)
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_MOVE"),true)
          @status = LightTool::MOVING
          view.invalidate
        elsif ph.test_point(@direction,x,y,$d5pixel)
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_EDIT"),true)
          @status = LightTool::EDITING
          view.invalidate
        else
          # exit tool
          Sketchup.active_model.select_tool(nil)
        end
      when LightTool::MOVING
        # offset = point-@center
        # @center = point
        # @direction=@direction+offset
        # new_trans = Geom::Transformation.new(offset)
        # $d5currentInst.transform! new_trans
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        point=@mouse_ip.position
        if @direction!=point
          updateCenter(point)
          @status = LightTool::STATIC
          view.invalidate
          Sketchup.active_model.commit_operation
        end
      when LightTool::EDITING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        point=@mouse_ip.position
        if @center!=point
          updateDirection(point)
          @status = LightTool::STATIC
          view.invalidate
          Sketchup.active_model.commit_operation
        end
      end
    end

    def onMouseMove(flags, x, y, view)
      case @status
      when LightTool::NONE
        @mouse_ip.pick(view,x,y)
        view.invalidate
      when LightTool::STATIC
        @mouse_ip.pick(view,x,y)
        view.invalidate
      when LightTool::MOVING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        if @mouse_ip.position!=@direction
          view.invalidate
        end
      when LightTool::EDITING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        if @mouse_ip.position!=@center
          view.invalidate
        end
      when LightTool::CREATING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        if @mouse_ip.position!=@center
          view.invalidate
        end
      end
    end

    def draw(view)
      view.tooltip=@mouse_ip.tooltip if @mouse_ip.valid?
      getDirection($d5currentInst)
      curmouse=@mouse_ip.position
      mm2inch = 0.0393701
      len = 750*mm2inch
      wid = 15*mm2inch
      pts=[Geom::Point3d.new(-len,-wid,0),Geom::Point3d.new(-len,wid,0),Geom::Point3d.new(len,wid,0),Geom::Point3d.new(len,-wid,0)]

      case @status
      when LightTool::NONE #新建状态下，选择底面中心中
        view.draw_points(@mouse_ip.position, $d5pixel, 2, $d5green)
      when LightTool::CREATING #新建状态下，选择方向标志点中
        view.draw_points(@center, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@center,curmouse)
        view.line_stipple = "_"
        view.draw_line([@center,curmouse])
        view.draw_points(curmouse, $d5pixel, 2, $d5green)
        trans = Geom::Transformation.new(@center,@center-curmouse)
        for point in pts
          point.transform! trans
        end
        view.drawing_color=$d5transparent
        view.draw(GL_QUADS,pts)
      when LightTool::STATIC #待编辑状态下，底面中心，方向标志点，四个边的顶点都需要画出来
        view.draw_points(@center, $d5pixel, 2, $d5blue)
        view.draw_line([@center,@direction])
        view.draw_points(@direction, $d5pixel, 2, $d5blue)
        # trans=$d5currentInst.transformation
        # for point in pts
        #   point.transform! trans
        # end
        # view.drawing_color=$d5transparent
        # view.draw(GL_QUADS,pts)
      when LightTool::EDITING #编辑状态下，移动direction
        view.draw_points(@center, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@center,curmouse)
        view.line_stipple = "_"
        view.draw_line([@center,curmouse])
        view.draw_points(curmouse,$d5pixel, 2, $d5green)

        trans=$d5currentInst.transformation
        curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
        backRotationTrans = curRotationTrans.inverse
        newRotationTrans = Geom::Transformation.new(@center,@center-curmouse)
        for point in pts
          point.transform! trans
          point.transform! backRotationTrans
          point.transform! newRotationTrans
        end

        view.drawing_color=$d5transparent
        view.draw(GL_QUADS,pts)
      when LightTool::MOVING #移动状态下，移动底面中心中
        view.draw_points(curmouse, $d5pixel, 2, $d5green)
        view.draw_points(@direction, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@direction,curmouse)
        view.line_stipple = "_"
        view.draw_line([curmouse,@direction])

        trans=$d5currentInst.transformation
        curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
        backRotationTrans = curRotationTrans.inverse
        newRotationTrans = Geom::Transformation.new(curmouse,curmouse-@direction)
        for point in pts
          point.transform! trans
          point.transform! backRotationTrans
          point.transform! newRotationTrans
        end

        view.drawing_color=$d5transparent
        view.draw(GL_QUADS,pts)
      else
        # puts "error in draw of StripTool"
      end
    end
  end

  class RectTool
    attr_reader :direction, :center, :mouse_ip
    def getDirection(inst)
      if inst==nil or inst.deleted?
        return
      end
      trans=$d5currentInst.transformation
      mm2inch = 0.0393701
      depth = 1500*mm2inch
      o = ORIGIN
      regular_direction = Geom::Point3d.new(0,0,-depth)
      @center = o.transform(trans)
      @direction = regular_direction.transform(trans)

      regular_edgePoints=[Geom::Point3d.new([-2,0,0]),Geom::Point3d.new([0,-2,0]),Geom::Point3d.new([2,0,0]),Geom::Point3d.new([0,2,0])]
      for i in 0..3
        @edgePoints[i]=regular_edgePoints[i].transform(trans)
      end
      # putKeyPoints(inst)
    end

    def initialize
      @status = LightTool::NONE
      @direction = nil
      @center = nil
      @edgePoints = Array.new(4)
      @vertices = Array.new(4)
      @mouse_ip = Sketchup::InputPoint.new
    end



    def activate
      if true==$d5doubleClick
        @status = LightTool::STATIC
        getDirection($d5currentInst)
        $d5doubleClick = false
      else
        $d5currentInst = nil
        @status = LightTool::NONE
      end
    end

    def deactivate(view)
      view.invalidate
    end
    def getExtents
      bb = Sketchup.active_model.bounds
      if @mouse_ip.valid?
        bb.add(@mouse_ip.position)
      end
      if @center!=nil
        bb.add(@center)
      end
      if @direction!=nil
        bb.add(@direction)
      end
      return bb
    end
    def resume(view)
      # puts "resume: view = #{view}"
      view.invalidate
    end

    def suspend(view)
      # puts "suspend: view = #{view}"
    end

    def onCancel(reason, view)
      # puts "MyTool was canceled for reason ##{reason} in view: #{view}"
      # exit tool
      Sketchup.active_model.select_tool(nil)
    end

    def addRectCompDef(view)

      # normal=@direction-@center
      # normal.normalize!
      # normal.x*=5
      # normal.y*=5
      # normal.z*=5
      # @direction=@center+normal
      #
      # ents = Sketchup.active_model.entities
      # rectCompDef = Sketchup.active_model.definitions.add "D5RenderLight.Rect"
      # rectEnts = rectCompDef.entities
      #
      # tempCircle = ents.add_circle(ORIGIN,normal,2)
      # @edgePoints=[tempCircle[0].vertices[0].position,tempCircle[6].vertices[0].position,tempCircle[12].vertices[0].position,tempCircle[18].vertices[0].position]
      # ents.erase_entities tempCircle

      #
      #
      # selfoffset=@edgePoints[0]-ORIGIN
      # vs = Array.new(4) #vertices
      # vs[0]=@edgePoints[1]+selfoffset
      # vs[1]=@edgePoints[1]-selfoffset
      # vs[2]=@edgePoints[3]-selfoffset
      # vs[3]=@edgePoints[3]+selfoffset
      # rectEnts.add_face(vs)
      # inst = ents.add_instance(rectCompDef,Geom::Transformation.new(@center))
      #


      ents = Sketchup.active_model.entities
      rectCompDef = Sketchup.active_model.definitions.add "D5RenderLight.Rect"
      LightTool.setType(rectCompDef, LightTool::RECT_TYPE)
      rectEnts = rectCompDef.entities

      mm2inch = 0.0393701
      sideLength = 1000*mm2inch
      depth = 1500*mm2inch
      len = sideLength/2
      pts=[[-len,-len,0],[-len,len,0],[len,len,0],[len,-len,0]]
      regular_edgePoints=[Geom::Point3d.new([-len,0,0]),Geom::Point3d.new([0,-len,0]),Geom::Point3d.new([len,0,0]),Geom::Point3d.new([0,len,0])]
      regular_direction = Geom::Point3d.new(0,0,-depth)
      rectEnts.add_face(pts)

      rectEnts.add_line(regular_edgePoints[0],regular_edgePoints[2])
      rectEnts.add_line(regular_edgePoints[1],regular_edgePoints[3])

      new_origin=@center
      new_zaxis=@center-@direction
      trans = Geom::Transformation.new(new_origin,new_zaxis)

      inst = LightTool.addInst rectCompDef,trans
      $d5currentInst = inst

      for i in 0..3
        # @vertices[i]=pts[i].transform(trans)
        @edgePoints[i]=regular_edgePoints[i].transform(trans)
      end
      @direction = regular_direction.transform(trans)

      @status = LightTool::STATIC
      view.invalidate
    end



    def updateDirection(new_direction) #更改朝向
=begin
      rectEnts=$d5currentInst.definition.entities

      normal=@direction-@center
      ents = Sketchup.active_model.entities


      rectEnts.clear!
      sizea=@edgePoints[0].distance(@edgePoints[2])
      sizeb=@edgePoints[1].distance(@edgePoints[3])
      tempCircle = ents.add_circle(ORIGIN,normal,2)
      @edgePoints=[tempCircle[0].vertices[0].position,tempCircle[6].vertices[0].position,tempCircle[12].vertices[0].position,tempCircle[18].vertices[0].position]
      ratioa = sizea/@edgePoints[0].distance(@edgePoints[2])
      ratiob = sizeb/@edgePoints[1].distance(@edgePoints[3])
      for i in 0..3
        if i%2==0
          @edgePoints[i].x*=ratioa
          @edgePoints[i].y*=ratioa
          @edgePoints[i].z*=ratioa
        else
          @edgePoints[i].x*=ratiob
          @edgePoints[i].y*=ratiob
          @edgePoints[i].z*=ratiob
        end
      end

      ents.erase_entities tempCircle
      selfoffset=@edgePoints[0]-ORIGIN
      vs = Array.new(4) #vertices
      vs[0]=@edgePoints[1]+selfoffset
      vs[1]=@edgePoints[1]-selfoffset
      vs[2]=@edgePoints[3]-selfoffset
      vs[3]=@edgePoints[3]+selfoffset

      rectEnts.add_face(vs)

      normal.normalize!
      normal.x*=5
      normal.y*=5
      normal.z*=5
      @direction=@center+normal
=end
      curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
      backRotationTrans = curRotationTrans.inverse
      newRotationTrans = Geom::Transformation.new(@center,@center-new_direction)
      $d5currentInst.transform! backRotationTrans
      $d5currentInst.transform! newRotationTrans

      mm2inch = 0.0393701
      depth = 1500*mm2inch
      regular_direction = Geom::Point3d.new(0,0,-depth)
      @direction = regular_direction.transform($d5currentInst.transformation)
    end

    def updateCenter(new_center)
      curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
      backRotationTrans = curRotationTrans.inverse
      newRotationTrans = Geom::Transformation.new(new_center,new_center-@direction)
      $d5currentInst.transform! backRotationTrans
      $d5currentInst.transform! newRotationTrans

      mm2inch = 0.0393701
      depth = 1500*mm2inch
      regular_direction = Geom::Point3d.new(0,0,-depth)
      @direction = regular_direction.transform($d5currentInst.transformation)
      @center = new_center
    end
    def resize #改变了某条边上的标志点，更改光源大小
      rectEnts=$d5currentInst.definition.entities

      selfoffset=@edgePoints[0]-ORIGIN
      vs = Array.new(4) #vertices
      vs[0]=@edgePoints[1]+selfoffset
      vs[1]=@edgePoints[1]-selfoffset
      vs[2]=@edgePoints[3]-selfoffset
      vs[3]=@edgePoints[3]+selfoffset
      # puts "@edgePoints"
      # puts @edgePoints
      # puts "vertices"
      # puts vs

      rectEnts.clear!
      rectEnts.add_face(vs)
    end

    def onLButtonDown(flags, x, y, view)

      case @status
      when LightTool::NONE
        @mouse_ip.pick(view,x,y)
        point=@mouse_ip.position
        Sketchup.active_model.start_operation(D5Localize.info("LIGHTTIP_RECT"),true)
        @center = point
        @status = LightTool::CREATING
      when LightTool::CREATING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        point=@mouse_ip.position
        if @center!=point
          @direction = point
          addRectCompDef(view)
          @status = LightTool::STATIC
          Sketchup.active_model.commit_operation
        end
      when LightTool::STATIC
        @mouse_ip.pick(view,x,y)
        point=@mouse_ip.position
        ph = view.pick_helper
        # These do not require init()
        if ph.test_point(@center,x,y,$d5pixel)
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_MOVE"),true)
          @status = LightTool::MOVING
          view.invalidate
        elsif ph.test_point(@direction,x,y,$d5pixel)
          Sketchup.active_model.start_operation(D5Localize.info("LIGHT_EDIT"),true)
          @status = LightTool::EDITING
          view.invalidate
        else
          # exit tool
          Sketchup.active_model.select_tool(nil)
        # else
        #   for i in 0..3
        #     offsetEdgePoint=@edgePoints[i]+(@center-ORIGIN)
        #     if ph.test_point(offsetEdgePoint,x,y,10)
        #       Sketchup.active_model.start_operation("改变大小",true)
        #       puts "#{offsetEdgePoint} is picked"
        #       @edgePoints[i]=0
        #       @status = LightTool::RESIZING
        #       putKeyPoints($d5currentInst)
        #       break
        #     end
        #   end
        end
      when LightTool::MOVING
        # offset = point-@center
        # @center = point
        # @direction=@direction+offset
        # new_trans = Geom::Transformation.new(offset)
        # $d5currentInst.transform! new_trans
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        point=@mouse_ip.position
        if @direction!=point
          updateCenter(point)
          @status = LightTool::STATIC
          view.invalidate
          Sketchup.active_model.commit_operation
        end
      when LightTool::EDITING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        point=@mouse_ip.position
        if @center!=point
          updateDirection(point)
          @status = LightTool::STATIC
          view.invalidate
          Sketchup.active_model.commit_operation
        end
      # when LightTool::RESIZING
      #   for i in 0..3
      #     if @edgePoints[i]==0
      #       # puts "Nil"
      #       @edgePoints[i]=calculateShadow(point,i)
      #       opposite=(i+2)%4
      #       @edgePoints[opposite]=ORIGIN+(ORIGIN-@edgePoints[i])
      #
      #       # x=(@edgePoints[opposite].x+@edgePoints[i].x)/2
      #       # y=(@edgePoints[opposite].y+@edgePoints[i].y)/2
      #       # z=(@edgePoints[opposite].z+@edgePoints[i].z)/2
      #       # newcenter = Geom::Point3d.new(x,y,z)
      #       # puts newcenter
      #       # offset = newcenter-ORIGIN
      #       # @edgePoints[i]-=offset
      #       # @edgePoints[opposite]-=offset
      #       # puts @edgePoints
      #       # puts @edgePoints
      #       # $d5currentInst.transform! Geom::Transformation.new(offset)
      #       resize
      #       @status = LightTool::STATIC
      #       view.invalidate
      #       Sketchup.active_model.commit_operation
      #       break
      #     end
      #   end
      end
    end

    def onMouseMove(flags, x, y, view)
      case @status
      when LightTool::NONE
        @mouse_ip.pick(view,x,y)
        view.invalidate
      when LightTool::STATIC
        @mouse_ip.pick(view,x,y)
        view.invalidate
      when LightTool::MOVING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@direction))
        if @mouse_ip.position!=@direction
          view.invalidate
        end
      when LightTool::EDITING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        if @mouse_ip.position!=@center
          view.invalidate
        end
      when LightTool::CREATING
        @mouse_ip.pick(view,x,y,Sketchup::InputPoint.new(@center))
        if @mouse_ip.position!=@center
          view.invalidate
        end
      end
    end

    def calculateShadow(point,i) #计算point在face上的投影点，编辑标志点时使用
      pos = point+(ORIGIN-@center)
      opposite=(i+2)%4
      # view.draw_points([pos,@edgePoints[opposite]],10,2,"blue")
      # 用勾股定理计算距离
      c = pos.distance(@edgePoints[opposite]).to_f
      b = pos.distance_to_line([ORIGIN,ORIGIN-@edgePoints[opposite]])
      a = Math.sqrt(c*c-b*b)
      # puts "#{a},#{b},#{c}"

      normal = ORIGIN-@edgePoints[opposite]
      ratio = a/normal.length.to_f
      normal.x*=ratio
      normal.y*=ratio
      normal.z*=ratio
      shadowPoint=@edgePoints[opposite]+normal
      return shadowPoint
    end
    def draw(view)
      view.tooltip=@mouse_ip.tooltip if @mouse_ip.valid?
      getDirection($d5currentInst)
      curmouse=@mouse_ip.position
      mm2inch = 0.0393701
      len = mm2inch*500
      pts=[Geom::Point3d.new(-len,-len,0),Geom::Point3d.new(-len,len,0),Geom::Point3d.new(len,len,0),Geom::Point3d.new(len,-len,0)]

      case @status
      when LightTool::NONE #新建状态下，选择底面中心中
        view.draw_points(@mouse_ip.position, $d5pixel, 2, $d5green)
      when LightTool::CREATING #新建状态下，选择方向标志点中
        view.draw_points(@center, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@center,curmouse)
        view.line_stipple = "_"
        view.draw_line([@center,curmouse])
        view.draw_points(curmouse, $d5pixel, 2, $d5green)
        trans = Geom::Transformation.new(@center,@center-curmouse)
        for point in pts
          point.transform! trans
        end
        view.drawing_color=$d5transparent
        view.draw(GL_QUADS,pts)
      when LightTool::STATIC #待编辑状态下，底面中心，方向标志点，四个边的顶点都需要画出来
        view.draw_points(@center, $d5pixel, 2, $d5blue)
        view.draw_line([@center,@direction])
        view.draw_points(@direction, $d5pixel, 2, $d5blue)
        # trans=$d5currentInst.transformation
        # for point in pts
        #   point.transform! trans
        # end
        # view.drawing_color=$d5transparent
        # view.draw(GL_QUADS,pts)
      when LightTool::EDITING #编辑状态下，移动direction
        view.draw_points(@center, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@center,curmouse)
        view.line_stipple = "_"
        view.draw_line([@center,curmouse])
        view.draw_points(curmouse, $d5pixel, 2, $d5green)

        trans=$d5currentInst.transformation
        curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
        backRotationTrans = curRotationTrans.inverse
        newRotationTrans = Geom::Transformation.new(@center,@center-curmouse)
        for point in pts
          point.transform! trans
          point.transform! backRotationTrans
          point.transform! newRotationTrans
        end

        view.drawing_color=$d5transparent
        view.draw(GL_QUADS,pts)
      when LightTool::MOVING #移动状态下，移动底面中心中
        view.draw_points(curmouse, $d5pixel, 2, $d5green)
        view.draw_points(@direction, $d5pixel, 2, $d5blue)
        view.set_color_from_line(@direction,curmouse)
        view.line_stipple = "_"
        view.draw_line([curmouse,@direction])

        trans=$d5currentInst.transformation
        curRotationTrans = Geom::Transformation.new(@center,@center-@direction)
        backRotationTrans = curRotationTrans.inverse
        newRotationTrans = Geom::Transformation.new(curmouse,curmouse-@direction)
        for point in pts
          point.transform! trans
          point.transform! backRotationTrans
          point.transform! newRotationTrans
        end

        view.drawing_color=$d5transparent
        view.draw(GL_QUADS,pts)
      else
        puts "error in draw of RectTool"
      end
=begin
      if curEdgePoints!=nil
        offset = curCenter-ORIGIN
        for i in 0..3
          if @edgePoints[i]==0
            curEdgePoints[i]=calculateShadow(@mouse_ip.position,i)+(@center-ORIGIN)
            # view.draw_points(@edgePoints[opposite]+normal,10,2,"blue")
            # view.draw_line([@edgePoints[opposite],pos,@edgePoints[opposite]+normal])
            # curEdgePoints[i]=@mouse_ip.position
          else
            curEdgePoints[i]=@edgePoints[i]+offset
          end
        end
        view.draw_points(curEdgePoints,10,2,"blue")
      end
=end
    end
  end




  # A utility method to activate the tool. This is defined here for easy
  # reuse as well as making it easier to debug while developing. If this code
  # was directly in the `add_item` block and you needed to make changes to
  # how you activate the tool then it would not take effect until you
  # restarted SketchUp due to the load guard.
  def self.activate_point_tool
    Sketchup.active_model.select_tool($d5pointTool)
  end
  def self.active_strip_tool
    Sketchup.active_model.select_tool($d5stripTool)
  end
  def self.active_rect_tool
    Sketchup.active_model.select_tool($d5rectTool)
  end
  def self.active_spot_tool
    Sketchup.active_model.select_tool($d5spotTool)
  end
  $d5pointTool = PointTool.new
  $d5stripTool = StripTool.new
  $d5rectTool = RectTool.new
  $d5spotTool = SpotTool.new
  # unless file_loaded?(__FILE__)
  #   menu = UI.menu('Plugins')
  #   menu.add_item('Light Tool: PointTool') do
  #     self.activate_point_tool
  #   end
  #   menu.add_item('Light Tool: SpotTool') do
  #     self.active_spot_tool
  #   end
  #   menu.add_item('Light Tool: RectTool') do
  #     self.active_rect_tool
  #   end
  #   menu.add_item('Light Tool: StripTool') do
  #     self.active_strip_tool
  #   end
  #   file_loaded(__FILE__)
  # end


end


