# frozen_string_literal: true


module D5Material
  def self.start_sync(model)
    $d5materials_observer.rebuild_mat_name_cache(model)
    model.materials.add_observer($d5materials_observer)
  end

  def self.stop_sync(model)
    model.materials.remove_observer($d5materials_observer)
    $d5materials_observer.clear_mat_name_cache
  end
end

class D5MaterialsObserver < Sketchup::MaterialsObserver
  def initialize
    super
    # 延迟调用
    @delay_excuter = D5DelayExcuter.new

    # 【临时方案】 由于现在发给D5的材质关键字还是材质名，所以如果是材质名改变，则需要找到所有的使用该材质的entity，并更新mesh
    @mat_name_cache = Hash.new # 材质名缓存
  end

  def rebuild_mat_name_cache(model)
    model.materials.each { |mat| @mat_name_cache[mat.entityID] = mat.name }
  end

  def clear_mat_name_cache
    @mat_name_cache.clear
  end

  def onMaterialChange(materials, material)
    if !$d5Converter_connectionStatus || material.deleted?
      return
    end

    # 拖动滑块编辑时，会不断触发。所以延时发送
    @delay_excuter.excute do
      D5Benchmark.bm("update material change") do
        #D5dllFunc::Connect.call()
        D5InfoTrans.add_material_to_map(material, true)
        #D5dllFunc::Start_send.call($d5Converter_modelPtr)
        #D5dllFunc::DisConnect.call()

        # 【临时方案】 由于现在发给D5的材质关键字还是材质名，所以如果是材质名改变，则需要找到所有的使用该材质的entity，并更新mesh
        if @mat_name_cache[material.entityID] != material.name
          Sketchup::active_model.definitions.each do |definition|
            definition.entities.each do |entity|
              if entity.is_a?(Sketchup::Face)
                if entity.material == material || entity.back_material == material
                  # 根据新的更新逻辑，最小更新单位为组件实例，所以面更新时需要更新所有！直接！包含这个面的组件，即这里的definition的所有instance
                  definition.instances.each { |inst| $d5Converter_uniqueAddElementIdMap[inst.persistent_id] = 1 }
                  break
                end
              elsif entity.is_a?(Sketchup::Drawingelement) # Image Group ComponentInstance
                if entity.material == material
                  $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 1
                end
              end
            end
          end
          Sketchup::active_model.entities.each do |entity|
            if entity.is_a?(Sketchup::Face)
              if entity.material == material || entity.back_material == material
                # 根据新的更新逻辑，最小更新单位为组件实例，所以面更新时需要更新所有！直接！包含这个面的组件。这里是model,用 0 表示是model
                $d5Converter_uniqueAddElementIdMap[0] = 1
              end
            elsif entity.is_a?(Sketchup::Drawingelement) # Image Group ComponentInstance
              if entity.material == material
                $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 1
              end
            end
          end

          @mat_name_cache[material.entityID] = material.name
        end
      end
    end
  end

  def onMaterialUndoRedo(materials, material)
    if !material.deleted?
      onMaterialChange(materials, material)
    end
  end
end
$d5materials_observer=D5MaterialsObserver.new