# test config
D5CFT_ENABLE = false

# necessary for Enumerable::to_set
require "set"

# 性能测试方法，在控制台打印执行耗时
require 'benchmark'
module D5Benchmark
  # 测试启用开关，true时检测代码块耗时并打印；false时直接执行代码块
  @@D5BM_ENABLE = false

  def D5Benchmark.set_enable(enable)
    @@D5BM_ENABLE = enable
  end

  # "label     : real_time"
  def D5Benchmark.bm(label = "")
    if @@D5BM_ENABLE
      puts Benchmark.measure(label){ yield }.format("%10n: %r") # %n: label; %r: real time
    else
      yield
    end
  end
end

class D5DelayExcuter
  # 延迟调用, 以解决短时间多次调用的问题，使用者需为每一个执行创建一个此对象
  # 实现方式：设置倒计时，在倒计时结束后发送，期间再次触发更新的话，则重置倒计时。
  def initialize
    super
    @timer_id = nil
  end

  def excute
    # 停止前一个定时任务
    if @timer_id != nil
      UI.stop_timer(@timer_id)
    end

    # 创建新的定时任务
    @timer_id = UI.start_timer(0.1, false) do # 15个灯光定义，每个定义24个引用实例，共360个灯，同步耗时<0.04s。所以这里超时时间设为0.1
      @timer_id == nil
      yield if block_given?
    end
  end
end

# 以下为有依赖关系的代码，需依照下面的顺序加载
#
#

module D5Strs
  # 插件相关目录
  D5RESOURCE_PATH = File.dirname(__FILE__ ) + '/'
  SU_INSTALL_PATH = Sketchup.find_support_file("", "")
end

module D5Localize
  @info = {} # Hash.new
  @error = {} # Hash.new
  def self.info(key)
    value = @info[key]
    if value.nil?
      value = "default info"
    end
    return value
  end
  def self.error(key)
    value = @error[key]
    if value.nil?
      value = "default error"
    end
    return value
  end

  # error_code 0-6 represent SUCCESS,NOT_INSTALLED,NOT_OPEN,UNSUPPORTED_VERSION,RENDER_BUSY,USER_CANCEL,UNKNOW_ERROR
  CONNECT_KEY=[
      "SUCCESS", # 0
      "NOT_INSTALLED",
      "NOT_OPEN",
      "UNSUPPORTED_VERSION",
      "RENDER_BUSY",
      "RENDER_BUSY_UPDATING",
      "USER_CANCEL",
      "UNKNOW_ERROR",
      "NETWORK_ERROR"
  ]
  def self.connect_enum_to_key(enum)
    enum_range = 0..8 # include 8
    if enum_range.include?(enum)
      return CONNECT_KEY[enum]
    end
  end

  def self.localfile_to_hash(filename)
    stringsfile = D5Strs::D5RESOURCE_PATH + Sketchup.get_locale() + filename # TODO: 字符串的链接用 ‘+’ 效率较低
    if(!File.exist?(stringsfile))
      stringsfile = D5Strs::D5RESOURCE_PATH+"en-US" + filename # 未适配的语言环境下使用英文资源
    end
    if(!File.exist?(stringsfile))
      UI.messagebox("Can't find localization strings for D5Render")
    end

    hash = Hash.new
    File.open(stringsfile) {|f|
      string = f.read # This returns a string even if the file is empty.
      string.gsub!(/\r\n?/, "\n")
      string.each_line do |line|
        line = line.gsub("\n","")
        key_value = line.split('==')
        hash[key_value[0]] = key_value[1]
      end
    }
    return hash
  end

  def self.init
    #从文件初始化string的key和value
    @info = self.localfile_to_hash("/D5Render.strings")
    #初始化要用到的报错string的值
    @error = self.localfile_to_hash("/D5ErrorText.strings")
  end
end

# 依赖 D5Prep、D5dllFunc
module D5Message
  #根据message内容显示warning
  def D5Message.show_my_warning(message)
    properties = {
        :dialog_title => "D5 Converter "+D5Converter::EXTENSION.version+" - Warning",
        :scrollable => false,
        :resizable => false,
        :width => 335,
        :height => 200
    }
    html = <<-HTML
  <html>
    <meta charset="utf-8">
    <style type = "text/css">
    .text-font{
      margin-top:20px;
      font-family:Microsoft YaHei;
      color:#000000;
      font-size:14px;
    }
    .button {
     margin-top:15px;
     width:120px;
     height:32px;
     border-width:0px;
     cursor:pointer;
     outline:none;
     font-family:Microsoft YaHei;
     color:black;
     font-size:14px;
    }
    .button:hover{
      background: #5599FF;
    }
    </style>
    <body>
      <p class="text-font">#{message}</p>
      <div style="text-align: center">
        <button class="button" onclick="window.sketchup.close()">OK</button>
      </div>
    </body>
  </html>
    HTML
    dialog = UI::HtmlDialog.new(properties)
    dialog.set_html(html)
    dialog.add_action_callback('close') { |action_context|
      dialog.close
    }
    dialog.center
    dialog.show_modal
  end

  #报错未安装，打开文件资源管理器提示选择安装目录
  # error_code NOT_INSTALLED
  def D5Message.show_installation_warning(message)
    properties = {
        :dialog_title => "D5 Converter "+D5Converter::EXTENSION.version+" - Warning",
        :scrollable => false,
        :resizable => false,
        :width => 335,
        :height => 200
    }
    html = <<-HTML
  <html>
<meta charset="utf-8">
    <style type = "text/css">
    .text-font{
      margin-top:20px;
      font-family:Microsoft YaHei;
      color:#000000;
      font-size:14px;
    }
    .button {
     margin-top:15px;
     width:120px;
     height:32px;
     border-width:0px;
     cursor:pointer;
     outline:none;
     font-family:Microsoft YaHei;
     color:black;
     font-size:14px;
    }
    .button:hover{
      background: #5599FF;
    }
    </style>
    <body>
      <p class="text-font">#{message}</p>
      <div style="text-align: right">
        <button class="button" onclick="window.sketchup.browse_for_path()">Select Folder</button>
        <button class="button" onclick="window.sketchup.close()">OK</button>
      </div>
    </body>
  </html>
    HTML

    dialog = UI::HtmlDialog.new(properties)
    dialog.set_html(html)
    dialog.add_action_callback('close') { |action_context|
      dialog.close
    }
    dialog.add_action_callback("browse_for_path") { | action_context |
      path = UI.select_directory(
          title: "Installation Folder of D5 Render"
      )
      if path !=nil
        d5RenderPath = path.to_s#.gsub("/","\\\\")
        pathStatus = D5dllFunc::Set_d5render_path.call(d5RenderPath.encode("utf-16le"))
        if pathStatus == 0
          show_my_warning D5Localize.error("SELECT_D5FOLDER")
        else
          D5Config.set_d5_render_folder(d5RenderPath)
          dialog.close
        end
      end
    }
    dialog.center
    dialog.show_modal
  end


  # D5风格的控制台输出
  D5_PUTS_HEAD = "[D5 Converter]"
  D5_PUTS_LEVEL = [" ","Warning: ","Error  : "]
  def D5Message.d5_puts(info = "", level = 0)
    puts "" << D5_PUTS_HEAD << D5_PUTS_LEVEL[level] << info
  end
end

module D5Config
  # 从配置文件中获取 d5安装路径
  def D5Config.d5_render_folder
    stringfile = D5Strs::D5RESOURCE_PATH+"/d5_render_folder.strings"
    string =""
    if File.file?(stringfile)
      File.open(stringfile){|f|
        string = f.read
      }
    end
    return string
  end

  # 向配置文件写入 d5安装路径
  def D5Config.set_d5_render_folder input
    stringfile = D5Strs::D5RESOURCE_PATH+"/d5_render_folder.strings"
    aFile = File.new(stringfile, "w+")
    if aFile
      aFile.syswrite(input)
    else
      #puts "Unable to open file!"
    end
  end

  # 从配置文件中获取 d5安装路径
  def D5Config.texture_folder
    stringfile = D5Strs::D5RESOURCE_PATH+"/d5_texture_folder.strings"
    string =""
    if File.file?(stringfile)
      File.open(stringfile){|f|
        string = f.read
      }
    end
    #puts "get local texture folder:",string
    return string
  end

  #写入…
  def D5Config.set_texture_folder input
    stringfile = D5Strs::D5RESOURCE_PATH+"/d5_texture_folder.strings"
    aFile = File.new(stringfile, "w+")
    if aFile
      aFile.syswrite(input)
    else
      #puts "Unable to open file!"
    end
  end
end

# some preparation for the start, such as file IO and version update
module D5Prep

  #initialize the so
  def D5Prep.load_so
    # require 'C:\Users\Administrator\Documents\_mydoc\ruby_c_extension_examples_SU2021\Release (2.5)\x64\SUEX_C.so'
    versionNum = Sketchup.version_number
    if (1700000000...1900000000) === versionNum
      require D5Strs::D5RESOURCE_PATH+'ruby22/SUEX_C.so'
    elsif (1900000000...2100000000)===versionNum
      require D5Strs::D5RESOURCE_PATH+'ruby25/SUEX_C.so'
    elsif 2100000000<versionNum
      require D5Strs::D5RESOURCE_PATH+'ruby27/SUEX_C.so'
    end
  end

  def D5Prep.show_my_folder_select(location)
    properties = {
        :dialog_title => "D5 Converter #{D5Converter::EXTENSION.version} - #{D5Localize.info('TOOLBAR_SETTING')}",
        :scrollable => false,
        :resizable => false,
        :width => 280,
        :height => 230
    }
    photoLocation = D5Strs::D5RESOURCE_PATH + "pictures/ic_more.png"
    photoReset = D5Strs::D5RESOURCE_PATH + "pictures/ic_reset.png"
    html = <<-HTML
  <html>
  <head>
  <meta charset="utf-8">
  <title>D5 render Setting</title>
  <style>
   .input{
      margin-left:16px;
      float:left;
      text-overflow: ellipsis;
      overflow: hidden;
      white-space: nowrap;
      border:1px solid #C0C0C0;/*设置边框1px，实线，边框线色为#a0b3d6 */
      width: 155px;
      min-height: 16px;
      font-size: 12px;
      font-family:Microsoft YaHei;
      padding: 4px;
      color: #333;
      outline:0; /* 当元素获得焦点的时候，焦点框为0.去掉虚线框（或高亮框）*/
    }
  /* 当输入框为空时，显示默认文字 */
   .input:empty::before{
      content:attr(placeholder); /*获取属性的值 attr（value）,可以获得属性值 */
      color: #999;
    }
     .title-font {
      font-family:Microsoft YaHei;
      color:#000000;
      font-size:14px;
      margin-left:16px;
      margin-top:34px;
    }
    .container{
        width: 100%;
    }
    .container::after{
        content: "";
        display: block;
        clear: both;
    }
    .button {
     margin-top:40px;
     width:112px;
     height:32px;
     border-width:0px;
     cursor:pointer;
     outline:none;
     font-family:Microsoft YaHei;
     color:black;
     font-size:14px;
    }
    .button:hover{
      background: #5599FF;
    }
    .more-button{
      margin-left:8px;
      float:left;
      width:30px;
      height:30px;
      background: url("#{photoLocation}") no-repeat;
      background-position:center;
      cursor:pointer;
      border-style:none;
    }
    .reset-button{
      float:left;
      width:30px;
      height:30px;
      background: url("#{photoReset}") no-repeat;
      background-position:center;
      cursor:pointer;
      border-style:none;
    }
  </style>
  </head>
  <body style="margin:0">
  <script type='text/javascript'>
    function changeContent(msg){
      var input = document.getElementById("dir_shower");
      input.setAttribute("placeholder",msg);
      input.setAttribute("title",msg);
    }
  </script>
  <div style="width:260px;height:40px">
    <p class="title-font">#{D5Localize.info('LABEL_TEXTURES_FOLDER')}</p>
    <div class="container">
      <div id="dir_shower" href=# title="#{location}" class="input" contenteditable="false" placeholder="#{location}"></div>
      <button class = "more-button" onclick="window.sketchup.findDir()"></button>
      <button class = "reset-button" onclick="window.sketchup.setDefault()"></button>
    </div>
    <button class = "button" style="margin-left: 16px" onclick="window.sketchup.save()">#{D5Localize.info('BUTTON_SAVE')}</button>
    <button class = "button" onclick="window.sketchup.close()">#{D5Localize.info('BUTTON_CANCEL')}</button>
  </div>
  </body>
  </html>
    HTML
    dialog = UI::HtmlDialog.new(properties)
    dialog.set_html(html)

    textures_dir = location;
    dialog.add_action_callback('close') { |action_context|
      dialog.close
    }
    dialog.add_action_callback('save') { |action_context|
      # UI.messagebox("next is save process")
      D5Config.set_texture_folder textures_dir
      dialog.close
    }
    #dir 需要写到配置文件中
    dialog.add_action_callback('findDir') { |action_context|
      textures_dir = UI.select_directory.to_s
      if textures_dir !=""
        dialog.execute_script("changeContent('#{textures_dir}')")
      end
    }
    dialog.add_action_callback('setDefault') { |action_context|
      textures_dir = ""
      dialog.execute_script("changeContent('#{textures_dir}')")
    }
    dialog.center
    dialog.show_modal
  end

end

module D5Update

  #get remote version and compare it with local version
  def self.str_str(haystack, needle)
    return 0 if needle.empty?
    haystack =~ /#{needle}/ || -1
  end

  #need to test the network ping
  def self.unzip_file zip_file,all_dir
    result = false
    result = system "7z x \"#{zip_file}\" -y -o\"#{all_dir}\""
    #result represent whether extract process is successful
  end


  #下载并解压remote_version版本
  @d5Converter_downloaded_timer_condition = false #used in func:download_extract
  def self.download_extract(remote_version)
    # UI.messagebox("next will update")
    #download the zip file meanwhile show the progress bar
    # https://d5-converter-test.oss-cn-beijing.aliyuncs.com/update/su/d5_su_v0.1.0.zip
    url = ""
    if D5Converter::ENVIRONMENT == "cn"
      url = "http://cn.cdn.d5converter.d5techs.com/update/su/d5_su_v"+remote_version+".zip"
    elsif D5Converter::ENVIRONMENT == "usa"
      url = "http://usa.d5converter.d5cdn.com/update/su/d5_su_v"+remote_version+".zip"
    else # test for everything else
      url = "http://test.cdn.d5converter.d5techs.com/update/su/d5_su_v"+remote_version+".zip"
    end

    filename = url[url.rindex('/')+1, url.length-1]
    fn = D5Strs::D5RESOURCE_PATH+filename.to_s

    request = Sketchup::Http::Request.new(url, Sketchup::Http::GET)
    # request.set_download_progress_callback do |current, total|
    #   if current == total
    #   end
    # end
    request.start do |req, resp|
      if request.status == Sketchup::Http::STATUS_SUCCESS
        zip_file = D5Strs::D5RESOURCE_PATH + filename.to_s
        all_dir = D5Strs::D5RESOURCE_PATH + "../"
        if File.exist? zip_file
          File.delete zip_file
        end
        File.write fn, resp.body, mode: 'wb'
        unzip_file(zip_file, all_dir)
        File.delete(zip_file)

        @result = RES_SUCCESS
      else
        @result = RES_FAILED
      end
      @d5Converter_downloaded_timer_condition = true
    end
  end

  # 分析版本字符串，并比较。true: 线上版本更高；false: 其他情况。
  def self.check_version_update local_version,remote_version
    localVersion = local_version.split('.')
    remoteVersion = remote_version.split('.')
    if remoteVersion[0].to_i > localVersion[0].to_i or remoteVersion[1].to_i > localVersion[1].to_i or remoteVersion[2].to_i > localVersion[2].to_i
      return true
    else
      return false
    end
  end

  # 显示升级提示对话框，并等待更新完成或取消
  def self.show_my_update(local_version, remote_version)
    properties = {
        :dialog_title => "D5 Converter "+local_version+" - New Version",
        :scrollable => false,
        :resizable => false,
        :width => 415,
        :height => 290
    }
    photoLocation = D5Strs::D5RESOURCE_PATH + "pictures/img_2.png"
    html = <<-HTML
  <html>
<meta charset="utf-8">
    <style type = "text/css">
    .button {
     width:182px;
     height:32px;
     border-width:0px;
     cursor:pointer;
     outline:none;
     font-family:Microsoft YaHei;
     color:black;
     font-size:14px;
    }
    .button:hover{
      background: #5599FF;
    }
    .title-font {
      font-family:Microsoft YaHei;
      color:#000000;
      font-size:20px;
    }
    .body-font {
      font-family:Microsoft YaHei;
      color:#000000;
      font-size:12px;
    }
    .link-font{
      font-family:Microsoft YaHei;
      color:#0000FF;
      font-size:12px
    }
    
    </style>
    </body>
    <body style="margin:0">
      <img src="#{photoLocation}" width="100%" />
      <h1 class="title-font" style="margin-left: 24px">
        New Version #{remote_version}
        <a href="#{D5Localize.info('WHATS_NEW')}" target="_blank" class="link-font">(What's new)</a>
      </h1>
      <div style="width:400px;height:14px">
        <p class="body-font" style="margin-left: 24px">Current Version #{local_version}</p>
      </div>
      <div style="width:400px;height:40px">
        <p class="body-font" style="margin-left: 24px">#{@remote_info["updateInfo"]}</p>
      </div>
      <div style="margin-bottom:12px;width:400px;height:32px">
        <button class = "button" style="margin-left: 16px" onclick="window.sketchup.update()">#{D5Localize.info("UPDATE_DOWNLOAD_NOW")}</button>
        <button class = "button" style="margin-right: 12px"onclick="window.sketchup.close()">#{D5Localize.info("UPDATE_REMIND_LATER")}</button>
      </div>
    </body>
  </html>
    HTML
    dialog = UI::HtmlDialog.new(properties)
    dialog.set_html(html)
    confirm_update = false
    dialog.add_action_callback('update'){ |action_context|
      confirm_update = true
      dialog.close
    }
    dialog.add_action_callback('close') { |action_context|
      dialog.close
    }
    dialog.center
    dialog.show_modal
    return confirm_update
  end

  def self.show_file_error_reinstall
    properties = {
        :dialog_title => "D5 Converter - Plugin File Error",
        :scrollable => false,
        :resizable => false,
        :width => 335,
        :height => 215
    }
    html = <<-HTML
<html>
  <meta charset="utf-8">
  <style type = "text/css">
    .button {
     margin-top:15px;
     width:130px;
     height:32px;
     border-width:0px;
     cursor:pointer;
     outline:none;
     font-family:Microsoft YaHei;
     color:black;
     font-size:14px;
    }
    .button:hover{
      background: #5599FF;
    }
    .text-font{
      margin-top:40px;
      margin-left: 12px;
      height:32px;
      font-family:Microsoft YaHei;
      color:#000000;
      font-size:14px;
    }  
  </style>
  <body style="margin:12px">
    <p class="text-font">#{D5Localize.info("UPDATE_FILE_CHECK_ERROR")}</p>
    <div style="margin-top:36px;margin-bottom:12px;width:307px;height:32px">
      <button class = "button" style="margin-left: 12px" onclick="window.sketchup.update()">#{D5Localize.info("UPDATE_DOWNLOAD_NOW")}</button>
      <button class = "button" style="margin-right: 12px"onclick="window.sketchup.close()">#{D5Localize.info("CANCEL_BUTTON")}</button>
    </div>
  </body>
</html>
    HTML
    dialog = UI::HtmlDialog.new(properties)
    dialog.set_html(html)
    confirm_update = false
    dialog.add_action_callback('update'){ |action_context|
      confirm_update = true
      dialog.close
    }
    dialog.add_action_callback('close') { |action_context|
      dialog.close
    }
    dialog.center
    dialog.show_modal
    return confirm_update
  end

  def self.goto_download
     `explorer.exe #{@remote_info["address"]}` # 等同于system ""
  end
  require D5Strs::D5RESOURCE_PATH+'json-2.5.1\lib\json.rb'
  def self.get_remote_info
    @get_remote_version_done = false

    url = ""
    if D5Converter::ENVIRONMENT == "cn"
      url = "https://cn.api.d5render.com/fusion-api/upgrade/pluginUpdateAvailable"
    elsif D5Converter::ENVIRONMENT == "usa"
      url = "https://usa.api.d5render.com/fusion-api/upgrade/pluginUpdateAvailable"
    else # test for everything else
      url = "https://test.api.d5render.com/fusion-api/upgrade/pluginUpdateAvailable"
    end
    # !!!request cannot be set to a temp variable
    @request = Sketchup::Http::Request.new(url, Sketchup::Http::POST)
    @request.headers["Content-Type"] = "application/json"
    @request.body = "{\"pluginType\": 3,\"version\": \"#{D5Converter::EXTENSION.version}\",\"cid\": \"#{D5GA.instance_variable_get(:@cid)}\" }"
    @request.start do |request, response|
      #puts request.status
      @remote_info = { "data"=>{"canUpgrade"=>false ,"address"=>"", "dcc"=>"", "updateInfo"=>"", "latestVersion"=>"0.0.0"} }
      if request.status == Sketchup::Http::STATUS_SUCCESS
        if response.body != ""
          json = response.body
          #puts url
          #puts @request.body
          #puts json
          begin
            j_return = JSON.parse(json)
            @remote_info = j_return["data"] if j_return["msg"]=="success"
          rescue
            puts "in rescue"
          end
          #puts @remote_info
        end
      else
        puts "D5 Converter: request fail in get_remote_version"
      end
      @get_remote_version_done = true
    end
  end

  def self.do_update
    @update_done = false

    get_remote_info
    # wait for getting remote version is done
    @version_timer = UI.start_timer(0.5, true){
      if @get_remote_version_done
        UI.stop_timer(@version_timer)

        # 检查版本，不是最新版则显示版本更新对话框
        local_ver = D5Converter::EXTENSION.version
        confirm_update = false
        if @remote_info["canUpgrade"]
          # 显示更新提示对话框，并返回用户是否确认选择更新
          confirm_update = show_my_update(local_ver, @remote_info["latestVersion"])
        end

        if confirm_update
          # 改为跳转到下载链接的方式
          goto_download

          @result = RES_CANCEL
          @update_done = true
        else
          # 取消更新或不需更新
          @result = RES_CANCEL
          @update_done = true
        end
      end
    }
  end

  # 更新完成标志
  @update_done = false
  def self.done?
    @update_done
  end

  # 更新执行结果
  RES_CANCEL  = 0
  RES_SUCCESS = 1
  RES_FAILED  = 2
  @result = RES_CANCEL # 0: 不需更新或用户取消；1: 更新成功；2: 更新失败
  def self.result
    @result
  end
end

module CHelper # TODO: 改名为D5CHelper
  #called by c api
  def self.getD5Pdll
    d5pPath = D5Strs::D5RESOURCE_PATH+"D5Plugin.dll"
    return d5pPath
  end

  def self.getPBdll
    pbPath = D5Strs::D5RESOURCE_PATH+"ProgressBar.dll"
    return pbPath
  end
end




$d5division_element2group_map = Hash.new
$d5division_group_change_map = Hash.new(0)
$d5division_status = false
$d5rootface_map = Hash.new
$d5rootface_map_old = Hash.new
#在这个module里进行模型的划分工作
module D5Divide
  def D5Divide.choose_send_type
    choice=D5Localize.info("SELECT")
    first_option_title=D5Localize.info("FIRST_OPTION_TITLE")
    first_option_dscp=D5Localize.info("FIRST_OPTION_DSCP")
    second_option_title=D5Localize.info("SECOND_OPTION_TITLE")
    second_option_dscp=D5Localize.info("SECOND_OPTION_DSCP")
    ok_button=D5Localize.info("OK_BUTTON")
    cancel_button=D5Localize.info("CANCEL_BUTTON")

    properties = {
        :dialog_title => "D5 Converter "+D5Converter::EXTENSION.version+" - Sync Mode",
        :scrollable => false,
        :resizable =>false,
        :width =>415,
        :height =>275
    }
    html = <<-HTML
<html>
<head>
  <meta charset="UTF-8">
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
	<style>
		* {
			transition: 0.2s;
		}
		.container {
			position: fixed;
			left: 0;
			top: 0;
			width: 400px;
			height: 238px;
			background-color: #F6F6F6;
			font-family: 'Segoe UI';
			user-select: none;
		}
		.header {
			position: fixed;
			left: 12px;
			top: 26px;
			font-size: 14px;
		}
		.option {
			position: fixed;
			left: 12px;
			width: 376px;
			height: 56px;
			border: 1px solid #CCCCCC;
			box-sizing: border-box;
		}
		.option:hover {
			border-color: #0078D7;
		}
		.option.active {
			border-color: #0078D7;
			background-color: rgb(229,241,251);
		}
		.option.active .check {
			display: flex;
		}
		.option.first {
			top: 54px;
		}
		.option.second {
			top: 114px;
		}
		.option .title {
			position: fixed;
			font-size: 16px;
		}
		.option.first .title {
			left: 24px;
			top: 58px;
		}
		.option.second .title {
			left: 24px;
			top: 118px;
		}
		.option .description {
			position: fixed;
			font-size: 12px;
		}
		.option.first .description {
			left: 24px;
			top: 88px;
		}
		.option.second .description {
			left: 24px;
			top: 148px;
		}
		.option .check {
			position: fixed;
			display: none;
			justify-content: center;
			align-items: center;
			width: 20px;
			height: 20px;
			left: 352px;
			border-radius: 50%;
			border: 1px solid #0078D7;
		}
		.option.first .check {
			top: 72px;
		}
		.option.second .check {
			top: 131px;
		}
		.check svg {
			width: 11px;
			height: 8px;
			fill: transparent;
			stroke: #0078D7;
		}
		.button {
			position: fixed;
			width: 120px;
			height: 33px;
			display: flex;
			justify-content: center;
			align-items: center;
			font-size: 14px;
			background-color: rgb(196,196,196);
		}
		#ok {
			left: 137px;
			top: 195px;
		}
		#cancel {
			left: 269px;
			top: 195px;
		}
		.button.disable {
			background-color: rgb(230, 230, 230);
			color: rgb(178, 178, 178);
		}
	</style>
</head>
<body>
  <div class="container">
		<div class="header">
			#{choice}
		</div>
		<div class="option first" onclick="onSelect(0)">
			<div class="title">#{first_option_title}</div>
			<div class="description">#{first_option_dscp}</div>
			<div class="check">
				<svg>
					<polyline points="0,5 4,8 11,0" />
				</svg>
			</div>
		</div>
		<div class="option second" onclick="onSelect(1)">
			<div class="title">#{second_option_title}</div>
			<div class="description">#{second_option_dscp}</div>
			<div class="check">
				<svg>
					<polyline points="0,5 4,8 11,0" />
				</svg>
			</div>
		</div>
		<div class="button disable" id="ok" onclick="onOk()">#{ok_button}</div>
		<div class="button" id="cancel" onclick="window.sketchup.cancel()">#{cancel_button}</div>
	</div>
	<script>
		var options = document.querySelectorAll('.option')
		var ok = document.querySelector('#ok')
		var cancel = document.querySelector('#cancel')
		var selected = -1;
		function onSelect (v) {
			if (selected === v) {
				selected = -1;
				options[v].classList.remove('active')
				ok.classList.add('disable')
			} else {
				selected = v;
				options[v].classList.add('active')
				options[1 - v].classList.remove('active')
				ok.classList.remove('disable')
			}
		}
		function onOk () {
			if (selected === -1) return;
      window.sketchup.ok(selected)
		}
	</script>
</body>
</html>
    HTML
    dialog = UI::HtmlDialog.new(properties)
    dialog.set_html(html)
    dialog.add_action_callback('ok') do|action_context,choice|
      $d5_send_model_type = choice+1
      dialog.close
    end
    dialog.add_action_callback('cancel') do |action_context|
      dialog.close
    end
    dialog.center
    dialog.show_modal
  end

  def D5Divide.division_pack_clear
    if $d5division_status==false
      return
    end
    path = Sketchup.active_model.path
    idx = path.rindex("\\")+1
    name = path[idx...path.length]
    division_xml="<root>\n"
    division_xml<<"<Group>
<GroupID>0</GroupID>
<ParentGroupID></ParentGroupID>
<GroupName>#{name}</GroupName>
<ShowHide>0</ShowHide>
</Group>\n\n"
#     $d5division_element2group_map.each do|key,value|
#       division_xml<<"<ElementNode>
# <ElementNodeID>#{key}</ElementNodeID>
# <ParentGroupID>#{value}</ParentGroupID>
# </ElementNode>\n\n"
#     end

    flag = false
    Sketchup.active_model.entities.each do |entity|
      if entity.is_a? Sketchup::Face or entity.is_a? Sketchup::Image
        $d5rootface_map[entity.persistent_id.to_s] = 0
        flag = true
      elsif entity.is_a?(Sketchup::ComponentInstance) or entity.is_a?(Sketchup::Group)
        key = entity.persistent_id.to_s
        showhide = 1 # 0:show 1:hide
        if entity.visible? and entity.layer.visible?
          showhide = 0
        end
        merge_group_name = entity.name
        if merge_group_name==""
          if entity.is_a?(Sketchup::Group)
            merge_group_name= D5Localize.info("GROUP_DEFAULT_NAME")
          elsif LightTool.getType(entity)!=nil #the instance is light
            continue
          else
            merge_group_name = entity.definition.name
          end
        end
        division_xml<<"<MergeGroup>
<MergeGroupID>#{key}</MergeGroupID>
<ParentGroupID>0</ParentGroupID>
<MergeGroupName>#{merge_group_name}</MergeGroupName>
<ShowHide>#{showhide}</ShowHide>
<Change>#{$d5division_group_change_map[key]}</Change>
</MergeGroup>\n\n" # 0:not change 1:change
      end
    end
# 最顶层的结构中存在不被包括的face
    if flag
      change_flag = 0
      if ($d5rootface_map.to_set-$d5rootface_map_old.to_set)!=nil or ($d5rootface_map_old.to_set-$d5rootface_map.to_set)!=nil
        change_flag = 1
      end
      $d5rootface_map_old=$d5rootface_map.clone()
      $d5rootface_map.clear()
      division_xml<<"<MergeGroup>
<MergeGroupID>000</MergeGroupID>
<ParentGroupID>0</ParentGroupID>
<MergeGroupName>RootFaceCollection</MergeGroupName>
<ShowHide>0</ShowHide>
<Change>#{change_flag}</Change>
</MergeGroup>\n\n"
    end
    division_xml<<"</root>\n"
    $d5division_group_change_map.clear()
    # $d5division_element2group_map.clear
    return division_xml
  end
end

class D5AppObserver <Sketchup::AppObserver
  def expectsStartupModelNotifications
    return true
  end
  def onOpenModel(model)
    # 激活灯光控制模块
    D5Light.initialize model

    if $d5Converter_connectionStatus==true
      $d5Converter_cmdImplement.stop
    end
  end
  def onNewModel(model)
    # 激活灯光控制模块
    D5Light.initialize model

    if $d5Converter_connectionStatus==true
      $d5Converter_cmdImplement.stop
    end
  end
end

module D5ViewAndScene
  def self.degree2Radians(degrees)
    radians = degrees * Math::PI / 180.0
    return radians
  end

  def self.radians2Degrees(radians)
    degrees = radians * 180.0 / Math::PI
    return degrees
  end

  UNIT_SCALE = 2.54
  def self.get_viewdata_fov(page_or_view)
    camera = page_or_view.camera

    # viewData
    viewData = Array.new
    viewData << camera.eye.to_a[1]*UNIT_SCALE
    viewData << camera.eye.to_a[0]*UNIT_SCALE
    viewData << camera.eye.to_a[2]*UNIT_SCALE
    viewData << camera.direction.to_a[1]
    viewData << camera.direction.to_a[0]
    viewData << camera.direction.to_a[2]
    viewData << camera.up.to_a[1]
    viewData << camera.up.to_a[0]
    viewData << camera.up.to_a[2]

    viewDataPtr = Fiddle::Pointer[viewData.pack('e*')]

    # fov
    if camera.perspective?
      # perspective and two-point perspective
      if camera.fov_is_height?
        # height fov(vFov) => width fov(hFov)
        aspectRatio = 1.78
        if page_or_view.is_a?(Sketchup::Page)
          aspectRatio = camera.aspect_ratio if camera.aspect_ratio!=0
        elsif page_or_view.is_a?(Sketchup::View)
          view = page_or_view
          topLeft = view.corner 0
          bottomRight = view.corner 3
          viewWidth = bottomRight.x - topLeft.x
          viewHeight = bottomRight.y - topLeft.y
          aspectRatio = viewWidth.to_f / viewHeight.to_f # aspectRatio = viewWidth/viewHeight = tan(hFOV/2) / tan(vFOV/2)
          # aspectRatio = ('%.2f' % (viewWidth.to_f/viewHeight.to_f)).to_f
        end

        vFov = camera.fov
        hFov = 2.0 * radians2Degrees(Math.atan(aspectRatio * Math.tan(degree2Radians(vFov/2.0))))
        fov = hFov # vFov < hFov ? vFov : hFov
      else
        fov = camera.fov
        # vFov = 2.0*radiansToDegrees(Math.atan(viewHeight.to_f/viewWidth.to_f*Math.tan(degreeToRadians(hFov/2.0))))
      end
    else
      fov = 0.0
    end

    # return
    [viewDataPtr, fov]
  end

  # 调用dll的函数添加并发送所有的场景
  def self.SendScenes()
    model = Sketchup.active_model
    pages = model.pages
    pages.each {|page|
      sceneName = page.name.to_s.encode("utf-16le")
      viewDataPtr, hFov = get_viewdata_fov(page)
      D5dllFunc::Add_scenes.call($d5Converter_modelPtr,sceneName,sceneName,viewDataPtr,hFov)
    }
    D5dllFunc::Start_sendScenes.call($d5Converter_modelPtr,model.path.to_s.encode("utf-16le"))
  end

  # 调用dll的函数发送当前视角的参数到d5render
  # $d5Redner_default_hFov = 90
  def self.sendCameraTransformFov(view)
    viewDataPtr, fov = get_viewdata_fov(view)
    D5dllFunc::Set_cameraTransformFov.call($d5Converter_modelPtr,viewDataPtr,fov)
  end

  #继承了ViewObserver，对View的events做出反应。当View发生改变时，重新发送camera的视角
  class D5ViewObserver < Sketchup::ViewObserver
    def initialize
      @last_send_time = nil
      super
    end
    def onViewChanged(view)
      @last_change_time=Time.now
      if @last_send_time!=nil and @last_change_time-@last_send_time<0.04
        return
      end
      @last_send_time=@last_change_time.clone
      #puts "send view in #{@last_change_time.hour}:#{@last_change_time.min}:#{@last_change_time.sec}.#{@last_change_time.usec}"
      D5ViewAndScene.sendCameraTransformFov(view)
      UI.start_timer(0.2) do # 0.04*5
        if @last_change_time!=@last_send_time and Time.now-@last_change_time>0.16 # 0.2-0.04
          #puts "send view in #{@last_change_time.hour}:#{@last_change_time.min}:#{@last_change_time.sec}.#{@last_change_time.usec} for the end"
          D5ViewAndScene.sendCameraTransformFov(view)
        end
      end
      # TODO: 新的实现方式。
      # def initialize
      # @view_cache = nil
      # @id = UI.start_timer(0.04, true) do
      #   if @view_cache
      #     time = Time.now
      #     puts time-@last_send_time if @last_send_time
      #     @last_send_time = time
      #     #puts "send view in #{@last_change_time.hour}:#{@last_change_time.min}:#{@last_change_time.sec}.#{@last_change_time.usec} for the end"
      #     puts Benchmark.measure { D5InfoTrans.sendCameraTransformFov(@view_cache) }
      #     @view_cache = nil
      #   end
      # end
    end
  end
  @view_observer = D5ViewObserver.new

  # record the status of live view
  @view_sync_status = false
  def self.view_sync_status
    @view_sync_status
  end

  def self.start_view_sync
    sendCameraTransformFov(Sketchup.active_model.active_view)
    # add observers
    Sketchup.active_model.active_view.add_observer(@view_observer)

    # changeViewOnStatus
    @view_sync_status = true
  end

  def self.stop_view_sync
    # remove observers
    Sketchup.active_model.active_view.remove_observer(@view_observer)
    # changeViewOnStatus
    @view_sync_status = false
  end
end

# 用于记录需要更新的element # TODO: $d5Converter_uniqueAddElementIdMap相关的移到D5InfoTrans
# map实际上是要实现set，map[id]==1的时候说明该ID存在于group中,map[id]==0的时候说明该ID是单独的face # TODO: 0 1的区分并没有实际使用
$d5Converter_uniqueAddElementIdMap = Hash.new

# 用于实现模型更新的Observer和相关方法；用于实现视角更新的Observer
# 1.对active_model.entities添加observer——实现对model层中每一个元素的【增加】和【修改】的更新 # TODO: 对组件的复制粘贴没有触发OnElementAdded
# 2.记录每一个组件定义的guid，guid会在组件内部更改后变化——实现对每一个组件内部元素【修改】的更新
module D5Observers
  # TODO: 待移除
  #辅助函数，用来在entities发生改变时递归添加所有的face
  #在Sketchup中用ruby实现对model.active_entites的dfs 把所有的face都找出来。最后把ID<==>类型的映射都保存在$d5Converter_uniqueAddElementIdMap中
  def D5Observers.addElementFromGropOrIns(entities)
    entities.each do |entity|
      if entity.is_a?Sketchup::Face
        $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 1
        entity.edges.each do |edge|
          edge.faces.each do |f|
            if f.entityID != entity.entityID
              $d5Converter_uniqueAddElementIdMap[f.persistent_id] = 1
            end
          end
        end
      elsif entity.is_a?Sketchup::Group
        D5Observers.addElementFromGropOrIns(entity.entities)
      elsif entity.is_a?Sketchup::ComponentInstance
        if LightTool.getType(entity.definition)==nil
          D5Observers.addElementFromGropOrIns(entity.definition.entities)
        end
      elsif entity.is_a?Sketchup::Image
        #TODO: don't know how to deal with
        $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 1
      end

    end
  end

  # TODO: $d5Converter_uniqueAddElementIdMap相关的移到D5InfoTrans
  # 缓存所有组件的GUID。组件修改后GUID会更新，以此为依据，在同步时判断每一个组件是否需要更新。
  # IN definitions OUT changed_def_array
  @@guid_cache = Hash.new
  def D5Observers.update_guid_cache(definitions)
    changed_def_array = Array.new
    definitions.each do |d|
      if @@guid_cache[d.entityID] != d.guid
        changed_def_array.push(d)
        @@guid_cache[d.entityID] = d.guid
      end
    end
    return changed_def_array
  end

  # TODO: $d5Converter_uniqueAddElementIdMap相关的移到D5InfoTrans
  def D5Observers.reset_guid_cache
    @@guid_cache.clear
  end

  class D5ModelObserver < Sketchup::ModelObserver # TODO: 用多继承，在Observer对象中增加一个控制变量，用来控制Observer是否需要触发。
    def initialize
      @transaction_id = 0
      super
    end
    attr_reader :transaction_id

    def onTransactionCommit(model)
      # #puts "onTransactionCommit: #{model}"
      @transaction_id=@transaction_id+1
    end
    def onTransactionRedo(model)
      # #puts "onTransactionRedo: #{model}"
      @transaction_id=@transaction_id+1
    end
    def onTransactionUndo(model)
      # #puts "onTransactionUndo: #{model}"
      @transaction_id=@transaction_id+1
    end
  end
  $d5mdl_observer = D5ModelObserver.new

  # 仅处理model.entities中的对象修改。继承了EntitiesObserver，监测Element的添加、删除、修改
  class D5EntitiesObserver < Sketchup::EntitiesObserver
    def onElementAdded(entities, entity)
      if entity.is_a?(Sketchup::Drawingelement)
        D5InfoTrans.set_need_update(entity)
      end

      if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance) || entity.is_a?(Sketchup::Image)
        onElementAdded(nil, entity.glued_to) if entity.respond_to?("glued_to") && entity.glued_to != nil # group,instance,image需要处理黏接的对象，见DC-198
      end
    end

    # 仅处理model.entities中的增删变化
    def onElementRemoved(entities, entity_id)
      if entities.parent.is_a?(Sketchup::ComponentDefinition)
        puts "[D5Converter] warning: not supposed to deal Definition entities in onElementRemoved"
        entities.parent.instances.each { |inst| $d5Converter_uniqueAddElementIdMap[inst.persistent_id] = 1 }
      elsif entities.parent.is_a?(Sketchup::Model)
        $d5Converter_uniqueAddElementIdMap[0] = 1
      end
      # #puts "onElementRemoved:#{entity_id}"
      # if $d5Converter_uniqueAddElementIdMap.has_key?(entity_id)
      #   if $d5Converter_uniqueAddElementIdMap[entity_id] == 0
      #     $d5Converter_uniqueAddElementIdMap.delete(entity_id)
      #   end
      # elsif $groupTree.has_key?(entity_id)
      #   $uniqueDeleteGroupIdMap[entity_id] = 0
      # else
      #   $uniqueDeleteElementIdMap[entity_id] = 0
      # end
    end

    def onElementModified(entities, entity)
      # #puts "onElementModified: #{entity.persistent_id}"
      if entity.is_a?(Sketchup::Drawingelement)
        D5InfoTrans.set_need_update(entity)
      end

      if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance) || entity.is_a?(Sketchup::Image)
        onElementAdded(nil, entity.glued_to) if entity.respond_to?("glued_to") && entity.glued_to != nil # group,instance,image需要处理黏接的对象，见DC-198
      end
    end
  end
  $d5Converter_MyEntitiesObserver = D5EntitiesObserver.new

  $d5Converter_layer_changed_map = Hash.new
  class D5LayersObserver < Sketchup::LayersObserver
    def onLayerChanged(layers, layer)
      # 只需考虑Instance或group的layer变化的情况
      $d5Converter_layer_changed_map[layer.entityID]=0
    end

    def onLayerFolderChanged(layers, layer_folder)
      layer_folder.each_layer { |layer| onLayerChanged(layers, layer) }
      layer_folder.each_folder { |folder| onLayerFolderChanged(layers, folder) }
    end
  end

end

require 'fileutils'

# record the status of connection. set true after sendmodel successfully
$d5Converter_connectionStatus = false
#设置两个tree是为了在更新的时候除去已经被删除的部分, value填入的是0，没有实际意义
$d5Converter_oldElementsTree = Hash.new
$d5Converter_newElementsTree = Hash.new
# material map存放已经添加过的material
$d5Converter_material_map = Hash.new

# five global variable initialized in InfoTrans.sendModel
# #used in addMeshAndInstance
# $d5Converter_uniqueFaceIdMap = Hash.new
# $d5Converter_curFaceNumber = 0
# $d5Converter_faceNumber = model.number_faces
# #used in processEntities
# $d5Converter_curEntityNumber = 0
# $d5Converter_EntitiesNumber = model.entities.size

# 处理模型/视角/entities update等信息在su和d5之间传输的时候需要处理的数据
# 在这个module里执行了大量的计算工作，大概要在这里重写一些函数 in C Api
module D5InfoTrans
  # 检查当前是否处于打开渲染器状态：状态为4代表RENDER_BUSY，正在渲染，返回true，否则返回false
  def self.checkD5RenderStatus()
    userCancelPtr = Fiddle::Pointer.new(0)

    # enum D5RenderStatus
    # {
    #     D5RenderNotOpen,
    #     D5RenderNotInProject,
    #     D5RenderNotReadyRendering,
    #     D5RenderNotReadyUpdating,
    #     D5RenderInProject,
    #
    #     D5RenderHardwareChecking = 10
    # };
    d5RenderStatus = D5dllFunc::GetD5RenderStatus.call(userCancelPtr)
    if d5RenderStatus != 4 # D5RenderInProject
      if d5RenderStatus == 0
        message = D5Localize.error("NOT_OPEN")
      elsif d5RenderStatus == 1
        message = D5Localize.error("NOT_IN_PROJECT")
      elsif d5RenderStatus == 2
        message = D5Localize.error("RENDER_BUSY")
      elsif d5RenderStatus == 3
        message = D5Localize.error("RENDER_BUSY_UPDATING")
      else #d5RenderStatus == 10 和 其他情况
        message = D5Localize.error("UNKNOW_ERROR")
      end
      D5Message.show_my_warning message

      if d5RenderStatus == 0 || d5RenderStatus == 1
        $d5Converter_connectionStatus = false # 断开连接状态
      end

      return false
    else
      return true
    end
  end


  # for c++ call. # TODO: don't know how to pass material, so pass the id instead. find_entity_by_id is very slow. move to CHelper
  def self.add_material_to_map_C(mat_id)
    material = Sketchup::active_model.find_entity_by_id(mat_id)
    if material!=nil && material.is_a?(Sketchup::Material)
      D5InfoTrans.add_material_to_map(material)
    else
      puts "add mat fail"
    end
  end

  # for c++ call. # TODO: don't know how to pass material, so pass the id instead. move to CHelper
  def self.generateMaterialName_C(mat_id)
    material = Sketchup::active_model.find_entity_by_id(mat_id)
    if material!=nil && material.is_a?(Sketchup::Material)
      D5InfoTrans.generateMaterialName(material)
    end
  end

  def self.int_to_hex a
    result = ""
    if a < 16
      result = result + "0" + a.to_s(16).upcase
    else
      result = result + a.to_s(16).upcase
    end
    return result
  end

  def self.color_to_hex a,b,c
    result = "#"
    result = result + D5InfoTrans.int_to_hex(a) + D5InfoTrans.int_to_hex(b) + D5InfoTrans.int_to_hex(c)
    return result
  end

  # 异常时返回“”。# TODO: 是否应该返回nil
  def self.get_texture_dir(model)
    #找到存放texture的本地文件夹，如果没有指定则默认为model所处的文件夹
    localFolder = D5Config.texture_folder
    resultDir = ""

    if localFolder == ""
      filePath = model.path.to_s
      if filePath != nil
        # defaultDir = filePath.split(".")[0]+"\\" #这样有风险，如文件名为abc.old.skp的文件。改为下面这样 ---- liuguanghao
        defaultDir = ""
        defaultDir << File.dirname(filePath) << "\\" << File.basename(filePath, ".*").strip # strip去掉最后可能出现的空格，解决fus-393
        defaultDir << "_d5c" # 文件夹名称加个后缀，避免文件夹名称以'.'或' '结尾。
        resultDir = defaultDir.gsub("\\","/")# 文件路径中的\替换为/   为什么要替换？
      end
    else
      resultDir = localFolder + "/"
    end

    if resultDir != ""
      FileUtils.makedirs(resultDir) unless File.exists?resultDir
    end

    return resultDir;
  end

  def self.set_need_update(entity)
    if entity.is_a?(Sketchup::Edge) || entity.is_a?(Sketchup::Face) # 特别的，对面的推拉操作，实际是对edge的修改，所以需要处理edge
      if entity.parent.is_a?(Sketchup::ComponentDefinition)
        entity.parent.instances.each { |inst| $d5Converter_uniqueAddElementIdMap[inst.persistent_id] = 1 }
      elsif entity.parent.is_a?(Sketchup::Model)
        $d5Converter_uniqueAddElementIdMap[0] = 1
      end
    elsif entity.is_a?(Sketchup::Group)
      $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 1
    elsif entity.is_a?(Sketchup::ComponentInstance)
      if LightTool.getType(entity.definition)==nil
        $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 1
      end
    elsif entity.is_a?(Sketchup::Image)
      $d5Converter_uniqueAddElementIdMap[entity.persistent_id] = 0
    elsif entity.is_a?(Sketchup::ComponentDefinition) # 处理 "设定为唯一" 时，组件内的组件内的图形pid变化，导致的更新异常
      entity.instances.each { |instance| $d5Converter_uniqueAddElementIdMap[instance.persistent_id] = 1 }
    elsif entity.is_a?(Sketchup::Entity) # 调整贴图UV后，会触发
      if entity.typename == "FaceTextureCoords" #处理贴图UV变化
        atr_dics = entity.parent
        face = atr_dics.parent if atr_dics
        if face.is_a?(Sketchup::Face)
          set_need_update(face)
        end
      end
    end
  end

  # 添加material到material map, 仅添加新增material
  def self.add_material_to_map(mat, force = false)
    if force || $d5Converter_material_map[mat.name]==nil
      inchToMScale = 0.0254

      materialName = mat.entityID.to_s

      color = [255,255,255,255]
      texture_dir = D5InfoTrans.get_texture_dir(Sketchup.active_model)
      texturePath = nil

      if mat.materialType == Sketchup::Material::MATERIAL_SOLID
        # 纯色，无纹理贴图
        color[0] = mat.color.red
        color[1] = mat.color.green
        color[2] = mat.color.blue
        color[3] = mat.color.alpha
      else #(mat.materialType == Sketchup::Material::MATERIAL_TEXTURED || mat.materialType == Sketchup::Material::MATERIAL_COLORIZED_TEXTURED)
        # 处理贴图
        if mat.texture != nil
          texturePath = File.join(texture_dir,mat.entityID.to_s+ "_" + File.basename(mat.texture.filename).to_s)
          mat.texture.write(texturePath, true) if mat.texture # true: Allows for the texture to be exported with the color adjustments.
        end
      end

      opacity = mat.alpha
      rough = opacity < 1 ? 0.0 : 0.4

      #以规定的xml格式存放material
      xml = "<Material>"
      xml << "<Name>"+materialName+"</Name>"
      xml << "<Type>Custom</Type>"
      xml << "<Opacity>"+opacity.to_s+"</Opacity>"
      xml << "<DiffuseColor>"+D5InfoTrans.color_to_hex(color[0], color[1], color[2])+"</DiffuseColor>"
      xml << "<EmissiveColor>#FFFFFF</EmissiveColor><EmissiveStrength>0</EmissiveStrength>"
      xml << "<TintColor>"+D5InfoTrans.color_to_hex(color[0], color[1], color[2])+"</TintColor>"
      xml << "<Roughness>"+rough.to_s+"</Roughness>"
      xml << "<BumpAmount>1.0</BumpAmount><Metallic>0</Metallic><Specular>0.30</Specular><IndexOfRefraction>1.52</IndexOfRefraction><IsSolidGlass>false</IsSolidGlass>"

      su_diffuse_textures = nil
      su_texture_width = nil
      su_texture_height = nil

      if texturePath == nil
        su_texture_width = "<TextureWidth>1.0</TextureWidth>"
        su_texture_height = "<TextureHeight>1.0</TextureHeight>"
        xml << su_texture_width+su_texture_height
      else
        curWidthStretch = mat.texture.width*inchToMScale
        curHeightStretch = mat.texture.height*inchToMScale
        su_texture_width = "<TextureWidth>"+curWidthStretch.to_s+"</TextureWidth>"
        xml << su_texture_width
        su_texture_height = "<TextureHeight>"+curHeightStretch.to_s+"</TextureHeight>"
        xml << su_texture_height
        su_diffuse_textures = "<DiffuseTexture>" + "<Filepath>" + texturePath + "</Filepath>"+"<Brightness>1.0</Brightness><IsInverted>false</IsInverted><UseExplicitTransformation>false</UseExplicitTransformation><Width>"+curWidthStretch.to_s+"</Width><Height>"+curHeightStretch.to_s+"</Height><U_Offset>0.0</U_Offset><V_Offset>0.0</V_Offset><Rotation>0.0</Rotation>"+"</DiffuseTexture>"
        xml << su_diffuse_textures
      end
      xml << "</Material>"

      # TODO: Read Enscape and D5 Materials
      #src_d5_xml = mat.get_attribute('D5.Material', 'MaterialData')
      #src_enscape_xml = mat.get_attribute('Enscape.Material', 'MaterialData')
      # if src_enscape_xml != nil || src_d5_xml!=nil
      #   if src_enscape_xml != nil
      #     final_xml = src_enscape_xml.to_s
      #   else
      #     final_xml = src_d5_xml.to_s
      #   end
      #   if match = final_xml.match(/<Opacity>[\s\S]*?<\/Opacity>/)
      #     final_xml[match.to_s] = su_opacity
      #   end
      #   if match = final_xml.match(/<DiffuseColor>[\s\S]*?<\/DiffuseColor>/)
      #     final_xml[match.to_s] = su_diffuse_color
      #   end
      #   if match = final_xml.match(/<TextureWidth>[\s\S]*?<\/TextureWidth>/)
      #     final_xml[match.to_s] = su_texture_width
      #   end
      #   if match = final_xml.match(/<TextureHeight>[\s\S]*?<\/TextureHeight>/)
      #     final_xml[match.to_s] = su_texture_height
      #   end
      #   if texturePath != nil
      #     if match = final_xml.match(/<DiffuseTexture>[\s\S]*?<\/DiffuseTexture>/)
      #       final_xml[match.to_s] = su_diffuse_textures
      #     end
      #   end
      #   # if match = final_xml.match(/<DiffuseTexture>[\s\S]*?<Filepath>(.+?)<\/Filepath>[\s\S]*?<\/DiffuseTexture>/)
      #   #   current_file_match = match.captures
      #   #   current_file_path = current_file_match[0].to_s
      #   #   if File.file?(current_file_path) != true
      #   #     if texturePath != nil
      #   #       # final_xml[current_file_path] = texturePath
      #   #       final_xml = xml
      #   #     end
      #   #   end
      #   # end
      # else
      #   final_xml = xml
      # end

      final_xml = xml
      D5dllFunc::Add_material.call($d5Converter_modelPtr,mat.name.encode("utf-16le"),final_xml.encode("utf-16le"))
      $d5Converter_material_map[mat.name] = 1
    end
  end

  @@def_instances_cache = Hash.new
  def self.update_def_instances_cache(definitions)
    definitions.each do |d|
      old_instances = @@def_instances_cache[d.entityID]
      new_instacnes = d.instances
      if old_instances != new_instacnes
        if $d5Converter_connectionStatus == true
          new_instacnes.each do |instance|
            if old_instances.nil? || !old_instances.include?(instance)
              # 新增的instance
              $d5Converter_uniqueAddElementIdMap[instance.persistent_id] = 1
            end
          end
        end
        @@def_instances_cache[d.entityID] = d.instances
      end
    end
  end

  def self.reset_def_instances_cache()
    @@def_instances_cache.clear
  end

  # 一种遍历所有entity的方法
  def self.eachDrawingElementInModel(model)
    model.entities.each { |entity| yield(entity) if entity.is_a?(Sketchup::Drawingelement) }
    model.definitions.each { |definition| definition.entities.each { |entity| yield(entity) if entity.is_a?(Sketchup::Drawingelement) } }
  end

  # 每一个instance的node。
  $d5Converter_NodeInstanceMap = Hash.new
  # 需要合并node的instance
  $d5Converter_mergeNodeInstance = Hash.new

  MAX_NODES_COUNT_LIMIT = 5000
  MAX_INSTANCES_USED_COUNT = 1000
  MAX_ENTITIES_IN_DEFINITION_COUNT = 200

  def self.get_node_instances(instance, array_node_instances)
    parent = instance.parent
    if (!parent.is_a?Sketchup::Model)
      parent.instances.each { |instance|
        if $d5Converter_mergeNodeInstance[instance]!=nil
          array_node_instances.push(instance)
        else
          get_node_instances(instance,array_node_instances)
        end
      }
    end
  end

  def self.collect_need_node_instances(definitions)
    definitions.each do |d|
      d.instances.each { |instance|
        array_node_instances = Array.new
        get_node_instances(instance,array_node_instances)
        if array_node_instances.empty?
          array_node_instances.push(instance)
        end
        $d5Converter_NodeInstanceMap[instance.persistent_id] = array_node_instances
      }
    end
  end

  def self.collect_merge_node_instances
    $d5Converter_mergeNodeInstance.clear

    # 限制node总量：方式一，从root开始遍历，累加每一层的instance总数，超出阈值时，这一层的所有instance标记为需要merge node
    level_entities = Sketchup::active_model.entities.to_a
    instances_count = level_entities.count

    while !level_entities.empty?
      new_entities_array = Array.new
      need_merge = false
      level_entities.keep_if { |entity| (entity.is_a?Sketchup::ComponentInstance) || (entity.is_a?Sketchup::Group) }

      level_entities.each do |entity|
        new_entities_array.concat entity.definition.entities.to_a
        need_merge = instances_count > MAX_NODES_COUNT_LIMIT
        instances_count+=(entity.definition.entities.select { |entity| (entity.is_a?Sketchup::ComponentInstance) || (entity.is_a?Sketchup::Group) }.count) if !need_merge
      end

      if need_merge
        level_entities.each { |entity| $d5Converter_mergeNodeInstance[entity.entityID]=0 }
        break
      else
        level_entities = new_entities_array
      end
    end
    #puts "instances_count:> "<<instances_count.to_s

    # 限制node总量：方式二，对所有definition按used_count由多到少排序，将used_count多的definition的所有instance的parent的instance 标记为需要merge node；
    # 直到used_count总数小于阈值。
    # NOT IMPLEMENTED!

    # Sketchup::active_model.definitions.each do |d|
    #   # 总量过多 TODO: 备用
    #   # too_many_used = d.count_used_instances > MAX_INSTANCES_USED_COUNT
    #   # if too_many_used
    #   #   definitions.instances.each { |instance|
    #   #     instance.parent.instances.each { |instance| $d5Converter_mergeNodeInstance[instance.entityID]=0 } unless instance.parent.is_a?Sketchup::Model
    #   #   }
    #   # end
    #
    #   # 限制每个definition中的entity数量（最好应该是限制instance数量）
    #   too_many_ent_in_definition = d.entities.count > MAX_ENTITIES_IN_DEFINITION_COUNT
    #   if too_many_ent_in_definition
    #     puts "merge node:" << d.name
    #     d.instances.each { |instance| $d5Converter_mergeNodeInstance[instance.entityID]=0 }
    #   end
    # end
  end

  def self.connect
    model = Sketchup.active_model
    D5dllFunc::Set_model_filepath.call($d5Converter_modelPtr, model.path.to_s.encode("utf-16le"))
    # log_to_file_rb.call()

    $d5_send_model_type = D5dllFunc::Get_send_model_type.call($d5Converter_modelPtr, model.path.to_s.encode("utf-16le"))
    if 2100000000>Sketchup.version_number
      $d5_send_model_type = 1
    end
    # puts $d5_send_model_type
    if $d5_send_model_type==0 # Unknown Type
      # TODO: split_model 暂不支持
      # D5Divide.choose_send_type
      $d5_send_model_type = 1
    end

    if $d5_send_model_type==1 #single_model
      $d5division_status = false
    elsif $d5_send_model_type==2 #split_model
      $d5division_status = true
    else #user cancel the transport
      return false
    end

    # 连接。失败时中断退出
    if $d5division_status
      connectInfo = D5dllFunc::Connect_split.call()
    else
      connectInfo = D5dllFunc::Connect.call($d5Converter_modelPtr)
    end
    if connectInfo != 0
      message = D5Localize.error(D5Localize::connect_enum_to_key(connectInfo))
      if message.nil?
        message = "other errors"
      end

      if connectInfo == 1
        D5Message.show_installation_warning message
      else
        D5Message.show_my_warning message
      end

      return false
    end
    return true
  end

  # 发送整个model，主要分为以下几部分：
  # 创立连接、写入material的texture、processEntities、根据connection发送optimized_model并且删除旧-新（：已经被删除的部分）、发送视角、断开连接
  # 其中需要改写的主要是processEntities，其余部分暂时不动
  def self.SendModel()
    model = Sketchup.active_model
    if model.path == ""
      D5Message.show_my_warning D5Localize.error("NOT_SAVE")
      return false
    end

    ## 连接
    connect_success = connect
    if !connect_success
      return false
    end

    ## 进度条开始
    $d5Converter_faceNumber = model.number_faces #faceNumber is used to calculater the percent of process bar # TODO: 进度条相关封装成类
    D5ProgressBar.start

    ## 预处理：收集变化信息等
    D5ProgressBar.update(0,"Preprocessing ...       ")# 用空格来对齐
    # 发送entities前，主动收集组件定义中需要更新的entities，到 $d5Converter_uniqueAddElementIdMap 中
    changed_defs_array = nil
    D5Benchmark.bm("collect changed instances") do
      changed_defs_array = D5Observers.update_guid_cache(model.definitions)
      if $d5Converter_connectionStatus == true
        # 临时方案：避免重复处理组件。 # 终极方案：重新设计增量更新的实现
        changed_defs_array.each do |definition|
          definition.instances.each { |instance| $d5Converter_uniqueAddElementIdMap[instance.persistent_id] = 1 }
        end
      end
    end
    # 收集新增的Instance
    D5Benchmark.bm("collect new instances") do
      update_def_instances_cache(model.definitions)
    end
    # 收集因为图层变化而需要更新的entities，到 $d5Converter_uniqueAddElementIdMap 中
    if $d5Converter_connectionStatus == true
      D5Benchmark.bm("collect layer changed instance and group") do
        eachDrawingElementInModel(model) do |entity|
          if $d5Converter_layer_changed_map.fetch(entity.layer.entityID, nil)!=nil
            set_need_update(entity)
          end
        end
        $d5Converter_layer_changed_map.clear
      end
    end
    # 收集需要作为node处理的组件实例，到 $d5Converter_mergeNodeInstance 中
    D5Benchmark.bm("collect need merge node instances") do
      collect_merge_node_instances
    end
    # 收集所有是node的instance
    D5Benchmark.bm("collect_need_node_instances") do
      $d5Converter_NodeInstanceMap.clear
      collect_need_node_instances(model.definitions)
    end
    # 收集由于需要node优化而需要更新的Instance
    D5Benchmark.bm("collect_need_update_node") do
      $d5Converter_uniqueAddElementIdMap.each do |instance_pid|
        node_instances = $d5Converter_NodeInstanceMap[instance_pid]
        node_instances.each { |node_instance| $d5Converter_uniqueAddElementIdMap[node_instance.persistent_id]=0 } if node_instances
        if node_instances.nil?
          D5Message.d5_puts("In preprocess set node_instance to add_map, no found instances in NodeInstanceMap",2)
        end
      end
    end

    ## 发送mesh和Instance
    if ($d5Converter_connectionStatus == false || !$d5Converter_uniqueAddElementIdMap.empty?)
      if 1700000000 <= Sketchup.version_number
        D5Benchmark.bm("process entities"){ SUEX_C.processEntitiesC() }
      end
      # delete the new-old(deleted entities in su) in d5
      if $d5Converter_connectionStatus == true
        # 已连接状态
        # 进度条。由于此步骤耗时较长，将此步骤的过程计算到进度条中。前面过程结束后，进度条应完成了90%。此步骤占总进度的10%。
        D5ProgressBar.update(90)

        # 更新模型，需要对比新旧，找到需要删除的
        D5Benchmark.bm("delete the new-old") do
          # 遍历所有old_tree中的项，判断new_tree中是否存在。不存在则视为需要删除。
          cur_index = 0
          total_count = $d5Converter_oldElementsTree.size
          update_step = total_count / 10
          update_step = 1 if update_step < 1 # 不能为0

          $d5Converter_oldElementsTree.each_key do |key|
            # 更新进度条
            cur_index = cur_index + 1
            if cur_index % update_step == 0
              percent = 90.0 + 10.0 * cur_index / total_count
              D5ProgressBar.update(percent.round)
            end

            if $d5Converter_newElementsTree.fetch(key, nil) == nil # 第二个参数是找不到key时的返回值。当其是用Hash.new方式创建时，$d5Converter_newElementsTree[key].nil? 与其效果相同
              if $d5division_status
                # 分组同步的删除
                D5dllFunc::Delete_element_split.call($d5Converter_modelPtr,key.to_s.encode("utf-16le"),$d5division_element2group_map[key].encode("utf-16le"),$d5MergeGroup_Ptr)
                merge_group_id= key.index('_')==nil ? nil : key[0...key.index('_')]
                if merge_group_id!=nil
                  $d5division_group_change_map[merge_group_id] = 1
                end
              else
                D5dllFunc::Delete_node.call($d5Converter_modelPtr,key.to_s.encode("utf-16le"))
              end
            end
          end
        end
      else
        # 未连接状态，进度条跳过此阶段
        D5ProgressBar.update(100)
      end
      $d5Converter_oldElementsTree = $d5Converter_newElementsTree
      $d5Converter_newElementsTree = Hash.new
    end

    ## 后处理
    # entities处理完成后，重置此缓存
    $d5Converter_uniqueAddElementIdMap.clear()

    ## D5Plugin发送数据到D5。发送数据这部分暂时认为不会失败。
    D5ProgressBar.update(100,"Almost done ...     ")# 用空格来对齐
    if $d5division_status
      division_xml = D5Divide.division_pack_clear()
      D5dllFunc::Optimize_model_split.call($d5Converter_modelPtr,division_xml.encode("utf-16le"),$d5MergeGroup_Ptr)
      D5dllFunc::Start_send_split.call($d5Converter_modelPtr)
      D5dllFunc::DisConnect_split.call()
    else
      D5dllFunc::Start_send.call($d5Converter_modelPtr)
      D5dllFunc::DisConnect.call($d5Converter_modelPtr)
    end

    ## 进度条结束
    D5ProgressBar.end
    return true
  end
end

# 对应UI的命令实现
class D5CmdImplement
  def tutorial
    # link = "http://cn.d5render.com/"
    link = D5Localize.info("TUTORIAL")
    UI.openURL(link)
  end

  def start
    model = Sketchup.active_model
    if model.path == ""
      D5Message.show_my_warning D5Localize.error("NOT_SAVE")
    elsif model.entities.count==0
      message=D5Localize.error("EMPTY")
      D5Message.show_my_warning message
    else
      render_folder = D5Config.d5_render_folder
      if render_folder != "" # 若已指定d5路径，则设置为指定的路径
        pathStatus = D5dllFunc::Set_d5render_path.call(D5Config.d5_render_folder.encode("utf-16le"))
        if pathStatus == 0 # 路径失效，导致设置失败，则清除已指定的路径
          D5Config.set_d5_render_folder("")
        end
      end

      connectInfo = D5dllFunc::StartLinkD5Render.call(model.path.to_s.encode("utf-16le"))
      #need to pop out a warning window
      if connectInfo != 0
        message = D5Localize.error(D5Localize::connect_enum_to_key(connectInfo))
        if message.nil?
          message = "other errors"
        end

        if connectInfo == 1
          D5Message.show_installation_warning message
        else
          D5Message.show_my_warning message
        end
      else
        if $d5division_status
          $d5MergeGroup_Ptr = Fiddle::Pointer.new(0)
          $d5MergeGroup_Ptr = D5dllFunc::CreateMergeGroupInfo_split.call()
        end
        $d5Converter_modelPtr = Fiddle::Pointer.new(0)
        $d5Converter_newModelPtr = Fiddle::Pointer.new(0)
        $d5Converter_modelPtr = D5dllFunc::Create_model.call()
        $d5Converter_material_map.clear
        send_success = false
        D5Benchmark::bm("Send Model") { send_success = D5InfoTrans.SendModel() }
        if send_success
          model.add_observer($d5mdl_observer)
          model.entities.add_observer($d5Converter_MyEntitiesObserver)
          model.definitions.add_observer($d5Converter_MyDefinitionsObserver)
          model.layers.add_observer(D5Observers::D5LayersObserver.new)
          $d5Converter_connectionStatus = true
          # view
          D5ViewAndScene.start_view_sync
          # lights
          D5Light.start_light_sync
          # materials update module
          D5Material.start_sync(model)
          # materials real-time update module.  -not finished
          #D5Material.initial_material(model)
        end
      end
    end
  end

  def stop
    # 缓存和状态重置
    $d5Converter_material_map.clear
    $d5Converter_connectionStatus = false
    $d5Converter_oldElementsTree = Hash.new
    D5Observers.reset_guid_cache()
    D5InfoTrans.reset_def_instances_cache()

    # stop light sync
    D5Light.stop_light_sync

    # view
    D5ViewAndScene.stop_view_sync

    # remove model sync observers
    model = Sketchup.active_model
    model.entities.remove_observer($d5Converter_MyEntitiesObserver)if $d5Converter_MyEntitiesObserver
    model.definitions.remove_observer($d5Converter_MyDefinitionsObserver)if $d5Converter_MyDefinitionsObserver

    # stop material sync
    D5Material.stop_sync(Sketchup.active_model)

    D5dllFunc::Destroy_model.call($d5Converter_modelPtr)
    if $d5division_status
      D5dllFunc::DestroyMergeGroupInfo_split.call($d5MergeGroup_Ptr)
    end
  end

  def model_sync
    if D5InfoTrans.checkD5RenderStatus()
      D5Benchmark::bm("Send Model") { D5InfoTrans.SendModel() }
    end
  end

  def view_switch(on)
    if on
      D5ViewAndScene.start_view_sync
    else
      D5ViewAndScene.stop_view_sync
    end
  end

  def send_scenes
    if D5InfoTrans.checkD5RenderStatus()
      D5ViewAndScene.SendScenes()
    end
  end

  def setting
    D5Prep.show_my_folder_select D5Config.texture_folder
  end

  # Note: SU data is modified and an operation will be committed.
  def light_hide(hide)
    op_name = (hide ? D5Localize.info("LIGHTTIP_HIDE") : D5Localize.info("LIGHTTIP_SHOW"))
    Sketchup::active_model.start_operation(op_name, true)
    for definition in Sketchup.active_model.definitions
      definition.instances.each { |inst| inst.hidden = hide } if LightTool.getType(definition)!=nil
    end
    Sketchup::active_model.commit_operation
  end
end

module D5CommandForTest
  @@timer
  def self.start
    SUEX_C.enable_test_port
    @@timer = UI.start_timer(0.1,true) do
      SUEX_C.run_test if SUEX_C.respond_to?("run_test")
    end
  end

  def self.stop
    UI.stop_timer(@@timer)
  end
end

# 修改代码后不重启sketchup就可以继续调试，只需重新load此文件
# 例：load "C:\\Users\\Administrator\\AppData\\Roaming\\SketchUp\\SketchUp 2020\\SketchUp\\Plugins\\su_to_d5render\\su_d5_light.rb"
D5Prep.load_so
require File.join(File.dirname(__FILE__),"su_d5_plugin_dll.rb")  # 加载依赖库，并初始化D5dllFunc中的函数常量
require File.join(File.dirname(__FILE__),"su_d5_light.rb") # require 只加载一次，load可以加载多次
require File.join(File.dirname(__FILE__),"su_d5_material.rb")
require File.join(File.dirname(__FILE__),"su_d5_ga_ui.rb")

module D5Main
  # init some used strings from configuration file
  D5Localize.init

  # 功能实现实例
  $d5Converter_cmdImplement = D5CmdImplement.new
  # 监控Sketch Up，用于在打开和新建文件时重置插件连接状态
  Sketchup.add_observer(D5AppObserver.new)

  # UI绑定响应，并显示工具栏和菜单入口
  D5PluginUI::bind_implement($d5Converter_cmdImplement)
  D5PluginUI::show

  # test 模块
  if D5CFT_ENABLE
    D5CommandForTest.start
  end

  # when open the Sketchup, check the version and decide whether need to update
  D5Update.do_update
  # wait for updating is done
  @update_timer = UI.start_timer(0.5, true){
    if D5Update.done?
      UI.stop_timer(@update_timer)

      # 检查so文件中记录的版本号，以解决文件替换失败的问题。期望此处理只需要在0.6.0升0.6.1版本时起作用，之后，通过安装包安装可以保证文件替换成功。
      so_version = (SUEX_C.respond_to?("version") ? SUEX_C.version : "0.0.0").split('.')
      rb_version = D5Converter::EXTENSION.version.split('.')
      if so_version[0].to_i != rb_version[0].to_i or so_version[1].to_i != rb_version[1].to_i or so_version[2].to_i != rb_version[2].to_i
        confirm_download = D5Update.show_file_error_reinstall
        D5Update.goto_download if confirm_download
      else
      end
    end
  }
end

# require 'rubygems/package'
# require 'zlib'
# require 'fiddle/import'
# require 'fiddle/types'
# module WinAPI
#   extend Fiddle::Importer
#   dlload 'kernel32.dll'
#   include Fiddle::Win32Types
#   extern 'int SetDllDirectory(LPCSTR)'
#   extern 'int GetDllDirectory(int,LPCSTR)'
# end

# 过期的或未实现的，没有实际被调用的方法
module D5InfoTrans_previous
  #Transform_to_s method is used to show the content of the transformation
  # 格式转换：把原字符串转换成每四个字符之间空格隔开，每四个后加一个换行符的16位字符串。形成4*4的matrix。
  # 在程序中未被调用？
  def self.transform_to_s t
    a = t.to_a.map {|e| e.to_s}
    s = a[0] + " " + a[1] + " " + a[2] + " " + a[3] + "\n" + a[4] + " " +a[5] + " " +a[6]+ " " +a[7] + "\n" + a[8] + " " +a[9] + " " +a[10]+ " " +a[11] + "\n" + a[12] + " " +a[13] + " " +a[14]+ " " +a[15] + "\n"
    return s
  end

  def self.generateMaterialName(material)
    name=""
    if material != nil
      name = material.name
      if material.texture!=nil
        name<<'('<<material.texture.filename<<')'
      else
        name<<'('<<"RGBA"<<material.color.to_a.to_s<<')'
      end
    end
    return name
  end


  def self.getDrawingElementsMaterial(entity)
    if entity.is_a?(Sketchup::Drawingelement)
      material = entity.material
      if entity.material==nil and entity.is_a?(Sketchup::Face)
        material = entity.back_material
      end
    end
    return material
  end

  def self.addImageAndVirtualFace(entity_pid, inputTrans,elementId,merge_group_id, fromC = false,entity_id = nil)
    if (1700000000...1800000000) === Sketchup.version_number  #su2017版本不支持image_rep
      return
    end

    if $d5division_status
      entity = Sketchup.active_model.find_entity_by_persistent_id entity_pid
    else
      entity = Sketchup.active_model.find_entity_by_id entity_id
    end

    image_path = entity.path
    vir_texture_name = File.basename(image_path).to_s

    vir_texture_width = entity.pixelwidth.to_m # = entity.image_rep.width.to_m # TODO: 这里像素转meter是否有意义
    vir_texture_height = entity.pixelheight.to_m# = entity.image_rep.height.to_m

    if (vir_texture_width != 0.0) && (vir_texture_height != 0.0) && image_path && (image_path != "")
      # 获取image_rep。# NOTE: su2021.1之前的版本，当图片数据异常时，entity.image_rep会导致su崩溃
      begin
        image_rep = entity.image_rep
      rescue
        puts "error! image is corrupt at: #{entity.origin}"
        return
      end
    else
      puts "error! image is corrupt at: #{entity.origin}"
      # 图片宽或高为0的异常情况；图片路径为空的异常情况
      return
    end

    if $d5Converter_material_map.has_key?(vir_texture_name)==false
      texture_dir = D5InfoTrans.get_texture_dir(Sketchup.active_model)
      vir_texture_path = File.join(texture_dir,entity.persistent_id.to_s+ "_" + vir_texture_name)

      begin
        image_rep.save_file(vir_texture_path)
      rescue
        # psd文件会保存失败
        puts "error! save image failed: #{vir_texture_path}"
      end

      image_xml="<Material>"
      image_xml<<"<Name>#{vir_texture_name}</Name>"
      image_xml<<"<Type>Custom</Type>"
      image_xml<<"<Opacity>1.0</Opacity>"
      image_xml<<"<DiffuseColor>#FFFFFF</DiffuseColor>"
      image_xml<<"<EmissiveColor>#FFFFFF</EmissiveColor><EmissiveStrength>0</EmissiveStrength><TintColor>#FFFFFF</TintColor><Roughness>0.4</Roughness><BumpAmount>1.0</BumpAmount><Metallic>0</Metallic><Specular>0.30</Specular><IndexOfRefraction>1.52</IndexOfRefraction><IsSolidGlass>false</IsSolidGlass>"
      image_xml<<"<TextureWidth>#{vir_texture_width}</TextureWidth>"
      image_xml<<"<TextureHeight>#{vir_texture_height}</TextureHeight>"
      image_xml<<"<DiffuseTexture><Filepath>#{vir_texture_path}</Filepath><Brightness>1.0</Brightness><IsInverted>false</IsInverted><UseExplicitTransformation>false</UseExplicitTransformation><Width>#{vir_texture_width}</Width><Height>#{vir_texture_height}</Height><U_Offset>0.0</U_Offset><V_Offset>0.0</V_Offset><Rotation>0.0</Rotation></DiffuseTexture>"
      image_xml<<"</Material>"
      D5dllFunc::Add_material.call($d5Converter_modelPtr,vir_texture_name.encode("utf-16le"),image_xml.encode("utf-16le"))
      $d5Converter_material_map[vir_texture_name] = 1
      # puts image_xml
    end

    # 构造虚拟面，用于显示图片。vir:virtual 虚拟face
    if $d5division_status
      entity_key = entity.persistent_id
    else
      entity_key = entity.entityID
    end
    if $d5Converter_uniqueFaceIdMap.has_key?(entity_key) == false
      vir_mesh = Array.new
      vir_normal=[entity.normal.y,entity.normal.x,entity.normal.z]*4
      vir_uv=[vir_texture_width,0,0,-vir_texture_height,0,0,vir_texture_width,-vir_texture_height]
      vir_index=[0,1,2,1,0,3]
      trans = entity.transformation
      vertices=[Geom::Point3d.new(image_rep.width,0,0).transform(trans),
                Geom::Point3d.new(0,image_rep.height,0).transform(trans),
                Geom::Point3d.new(0,0,0).transform(trans),
                Geom::Point3d.new(image_rep.width,image_rep.height,0).transform(trans)]
      for i in 0..3
        vir_mesh << vertices[i].y.to_cm
        vir_mesh << vertices[i].x.to_cm
        vir_mesh << vertices[i].z.to_cm
      end
      #call dll function
      meshP = Fiddle::Pointer[vir_mesh.pack('e*')]
      normalP = Fiddle::Pointer[vir_normal.pack('e*')]
      uvP = Fiddle::Pointer[vir_uv.pack('e*')]
      indexP = Fiddle::Pointer[vir_index.pack('i*')]
      # puts "virtual_face_persistent_id:#{entity.persistent_id}  mesh:#{vir_mesh}\nnormalData:#{vir_normal}\nuvData:#{vir_uv}\nindexData:#{vir_index}\n"
      D5dllFunc::Add_mesh.call($d5Converter_modelPtr,entity_key.to_s.encode("utf-16le"),meshP,12,normalP,12,uvP,8,nil,0,indexP,6)
      $d5Converter_uniqueFaceIdMap[entity_key] = 0
    end
    # transArray = inputTrans
    # transArrayPtr = Fiddle::Pointer[transArray.pack('e*')]
    D5dllFunc::Begin_element.call($d5Converter_modelPtr,elementId.encode("utf-16le"))
    transArrayPtr = Fiddle::Pointer[inputTrans.pack('e*')]
    if $d5division_status
      D5dllFunc::Add_instance_split.call($d5Converter_modelPtr,entity.persistent_id.to_s.encode("utf-16le"),vir_texture_name.to_s.encode("utf-16le"),transArrayPtr)
    else
      D5dllFunc::Add_instance.call($d5Converter_modelPtr,entity.entityID.to_s.encode("utf-16le"),vir_texture_name.to_s.encode("utf-16le"),transArrayPtr)
    end
    D5dllFunc::End_element.call($d5Converter_modelPtr)
    if $d5division_status
      D5dllFunc::Add_element_to_mergegroup_split.call($d5Converter_modelPtr,elementId.encode("utf-16le"), merge_group_id.encode("utf-16le"),$d5MergeGroup_Ptr)
    end

  end

  # 函数中有array和hash的数据读写操作，maybe changed to C Api
  # 另：函数中调用了dll文件，需要查询在c++中使用外部dll中的函数的方法
  def self.addMeshAndInstance(entity, inputTrans, parentMat, elementId,merge_group_id)
    # #puts "entityID:",entity.entityID
    # #puts "entityType:",entity.typename
    # if entity.material!=nil
    #   #puts "material name:",entity.material.name
    # else
    #   #puts "no material"
    # end
    # if parentMat!=nil
    #   #puts "parent material name:",parentMat.name
    # else
    #   #puts "no parent material"
    # end
    # #puts "elementId:",elementId
    # #puts '*************'

    # 处理material
    inchToMScale = 0.0254
    material_key = ""
    materialWidth = 1.0*inchToMScale
    materialHeight = 1.0*inchToMScale #these two values is to calculate the arrording uv by scretch
    front = false
    material = nil
    if entity.material != nil
      material = entity.material
      front = true
    else
      #if front has no material, check the back material
      material = entity.back_material
    end
    if material != nil
      material_key = material.name
      if material.texture != nil
        materialWidth = material.texture.width*materialWidth
        materialHeight = material.texture.height*materialHeight
      end
    end

    if material != nil
      D5InfoTrans.add_material_to_map(material)
    elsif parentMat != nil
      D5InfoTrans.add_material_to_map(parentMat)
    end

    #if the designer paint face without entering into group or instance, check the parent's material
    isUseParentMat = false
    if material == nil and parentMat != nil
      material_key = parentMat.name
      isUseParentMat = true
    end

    if $d5division_status
      mesh_key = entity.persistent_id.to_s
    else
      mesh_key = entity.entityID.to_s
    end

    # 一个face对应一个mesh，用$d5Converter_uniqueFaceIdMap避免重复创建mesh
    if $d5Converter_uniqueFaceIdMap.has_key?(mesh_key) == false
      mesh = entity.mesh(7)

      meshData = Array.new
      normalData = Array.new
      uvData = Array.new
      indexData = Array.new

      # if(elementId=="25924_20854")
      # #puts mesh.count_points
      # #puts mesh.polygons.size
      # end
      #process vertices,normals,uvs
      for i in 1..mesh.count_points
        r = mesh.point_at(i)
        t = mesh.normal_at(i)
        #need to reverse normalData to make light true
        meshData << r.x.to_cm
        meshData << r.y.to_cm
        meshData << r.z.to_cm

        normalData << t.x
        normalData << t.y
        normalData << t.z

        s = mesh.uv_at(i,front)
        if isUseParentMat == true and parentMat.texture != nil
          uvData << s.x * inchToMScale /s.z
          uvData << (0-s.y) * inchToMScale /s.z
        else
          #coordinate system difference between su and render
          uvData << s.x * materialWidth /s.z
          uvData << (0-s.y) * materialHeight /s.z
        end
      end
      #process index or the sequence of triangle drawing
      mesh.polygons.each do |polygon|
        indexData << polygon[0].abs - 1
        indexData << polygon[1].abs - 1
        indexData << polygon[2].abs - 1
      end
      # if uvData.size == 0
      #   mesh.uvs(false).each do |uv|
      #     t = uv.to_a
      #     uvData << t[0]
      #     uvData << t[1]
      #   end
      # end

      #call dll function
      meshDataPtr = Fiddle::Pointer[meshData.pack('e*')]
      normalDataPtr = Fiddle::Pointer[normalData.pack('e*')]
      uvDataPtr = Fiddle::Pointer[uvData.pack('e*')]
      indexDataPtr = Fiddle::Pointer[indexData.pack('i*')]
      D5dllFunc::Add_mesh.call($d5Converter_modelPtr,mesh_key.encode("utf-16le"),meshDataPtr,meshData.size,normalDataPtr,normalData.size,uvDataPtr,uvData.size,nil,0,indexDataPtr,indexData.size)
      $d5Converter_uniqueFaceIdMap[mesh_key] = 0
    end
    transArray = inputTrans
    transArrayPtr = Fiddle::Pointer[transArray.pack('e*')]

    # 当begin_elements时，如果已存在则会先删掉，所以不需要主动删除
    D5dllFunc::Begin_element.call($d5Converter_modelPtr,elementId.encode("utf-16le"))
    if $d5division_status
      D5dllFunc::Add_instance_split.call($d5Converter_modelPtr,mesh_key.encode("utf-16le"),material_key.encode("utf-16le"),transArrayPtr)
    else
      D5dllFunc::Add_instance.call($d5Converter_modelPtr,mesh_key.encode("utf-16le"),material_key.encode("utf-16le"),transArrayPtr)
    end
    D5dllFunc::End_element.call($d5Converter_modelPtr)
    if $d5division_status
      D5dllFunc::Add_element_to_mergegroup_split.call($d5Converter_modelPtr,elementId.encode("utf-16le"), merge_group_id.encode("utf-16le"),$d5MergeGroup_Ptr)
    end
  end

  #矩阵相乘，可能较为耗时
  #four_matrix_four method is used to deal with 4*4 matrix multiply
  def self.four_matrix_four a,b
    #need to change the origin position from inch to cm
    #*2.54 because the former format is inch we need cm
    scale = 2.54
    result = Array.new
    result << a[0]*b[0]+a[4]*b[1]+a[8]*b[2]+a[12]*b[3]
    result << a[1]*b[0]+a[5]*b[1]+a[9]*b[2]+a[13]*b[3]
    result << a[2]*b[0]+a[6]*b[1]+a[10]*b[2]+a[14]*b[3]
    result << a[3]*b[0]+a[7]*b[1]+a[11]*b[2]+a[15]*b[3]
    result << a[0]*b[4]+a[4]*b[5]+a[8]*b[6]+a[12]*b[7]
    result << a[1]*b[4]+a[5]*b[5]+a[9]*b[6]+a[13]*b[7]
    result << a[2]*b[4]+a[6]*b[5]+a[10]*b[6]+a[14]*b[7]
    result << a[3]*b[4]+a[7]*b[5]+a[11]*b[6]+a[15]*b[7]
    result << a[0]*b[8]+a[4]*b[9]+a[8]*b[10]+a[12]*b[11]
    result << a[1]*b[8]+a[5]*b[9]+a[9]*b[10]+a[13]*b[11]
    result << a[2]*b[8]+a[6]*b[9]+a[10]*b[10]+a[14]*b[11]
    result << a[3]*b[8]+a[7]*b[9]+a[11]*b[10]+a[15]*b[11]
    # result << a[0]*b[12]+a[4]*b[13]+a[8]*b[14]+a[12]*b[15]
    # result << a[1]*b[12]+a[5]*b[13]+a[9]*b[14]+a[13]*b[15]
    # result << a[2]*b[12]+a[6]*b[13]+a[10]*b[14]+a[14]*b[15]
    # result << a[3]*b[12]+a[7]*b[13]+a[11]*b[14]+a[15]*b[15]
    result << a[0]*b[12]*scale+a[4]*b[13]*scale+a[8]*b[14]*scale+a[12]*b[15]
    result << a[1]*b[12]*scale+a[5]*b[13]*scale+a[9]*b[14]*scale+a[13]*b[15]
    result << a[2]*b[12]*scale+a[6]*b[13]*scale+a[10]*b[14]*scale+a[14]*b[15]
    result << a[3]*b[12]*scale+a[7]*b[13]*scale+a[11]*b[14]*scale+a[15]*b[15]
    return result
  end

  # 递归可能较为耗时
  # called in sendModel(): D5InfoTrans.processEntities(entities, Geom::Transformation.new.to_a, nil, nil, true)
  # Geom::Transformation.new.to_a is a 4*4 identity matrix
  # [1.0, 0.0, 0.0, 0.0,
  #  0.0, 1.0, 0.0, 0.0,
  #  0.0, 0.0, 1.0, 0.0,
  #  0.0, 0.0, 0.0, 1.0]
  def self.processEntities(entities, inputTrans, parentMat, parentId, processBar,parentVisible = true,root = true)
    currentParentId = parentId
    entities.each do |entity|
      # $fileTree[entity.entityID] = entity.parent.entityID
      if entity.is_a?Sketchup::Face
        merge_group_id="000" #root group下直接的未被归类的face对应的根mergegroup
        if currentParentId!=nil
          # elementId = parentId.rpartition('_').last.to_s+'_'+entity.entityID.to_s
          elementId = currentParentId.to_s+'_'+entity.persistent_id.to_s
          merge_group_id = elementId[0...elementId.index('_')]
        else
          elementId = entity.persistent_id.to_s
        end
        if $d5Converter_uniqueAddElementIdMap.has_key?(entity.persistent_id) or !$d5Converter_oldElementsTree.has_key?(elementId) or !$d5Converter_connectionStatus
          $d5division_group_change_map[merge_group_id] = 1
        end

        if parentVisible and entity.visible? and entity.layer.visible?
          $d5division_element2group_map[elementId] = merge_group_id
          $d5Converter_newElementsTree[elementId] = 0

          # 需要重新发送mesh和Instance数据的情况：首次联动 or 待更新表中有此entity or 上一次的element中没有此entity（一个entity一个element）
          if !$d5Converter_connectionStatus || $d5Converter_uniqueAddElementIdMap.has_key?(entity.persistent_id) || !$d5Converter_oldElementsTree.has_key?(elementId)
            D5InfoTrans.addMeshAndInstance(entity,inputTrans,parentMat,elementId,merge_group_id)
            # if not $d5Converter_oldElementsTree.has_key?(entity.entityID)
          end
        end

        CHelper.updateBarInAddMesh()
        #maybe solve the thread problem
        #sleep(0.01)
      elsif entity.is_a?Sketchup::Image
        # construct a virtual face, change the image to a texture, and put it to the face.
        merge_group_id="000" #root group下直接的未被归类的face对应的根mergegroup
        if currentParentId!=nil
          elementId = currentParentId.to_s+'_'+entity.persistent_id.to_s
          merge_group_id = elementId[0...elementId.index('_')]
        else
          elementId = entity.persistent_id.to_s
        end
        if $d5Converter_uniqueAddElementIdMap.has_key?(entity.persistent_id) or !$d5Converter_oldElementsTree.has_key?(elementId) or !$d5Converter_connectionStatus
          $d5division_group_change_map[merge_group_id] = 1
        end

        if parentVisible and entity.visible? and entity.layer.visible?
          $d5division_element2group_map[elementId] = merge_group_id
          $d5Converter_newElementsTree[elementId] = 0

          # 需要重新发送mesh和Instance数据的情况：首次联动 or 待更新表中有此entity or 上一次的element中没有此entity（一个entity一个element）
          if !$d5Converter_connectionStatus || $d5Converter_uniqueAddElementIdMap.has_key?(entity.persistent_id) || !$d5Converter_oldElementsTree.has_key?(elementId)
            self.addImageAndVirtualFace(entity.persistent_id,inputTrans,elementId,merge_group_id,false,entity.entityID)
          end
        end

      elsif entity.is_a?Sketchup::Group
        curMat = nil
        if entity.material != nil
          curMat = entity.material
        end
        if curMat == nil
          curMat = parentMat
        end
        # $groupTree[entity.entityID] = 0
        if currentParentId!=nil
          # parentId = parentId.to_s+"_"+entity.entityID.to_s
          parentId = currentParentId.to_s+"_"+entity.persistent_id.to_s
        else
          # parentId = entity.entityID.to_s
          parentId = entity.persistent_id.to_s
        end
        if root #最顶层的mergegroup结构
          flag = true
        else
          flag = (parentVisible and entity.visible? and entity.layer.visible?)
        end
        processEntities(entity.entities,four_matrix_four(inputTrans,entity.transformation.to_a),curMat,parentId,false,flag,false)
      elsif entity.is_a?Sketchup::ComponentInstance
        if LightTool.getType(entity.definition)==nil
          curMat = nil
          if entity.material != nil
            curMat = entity.material
          end
          if curMat == nil
            curMat = parentMat
          end
          # $groupTree[entity.entityID] = 0
          if currentParentId!=nil
            parentId = currentParentId.to_s+"_"+entity.persistent_id.to_s
          else
            # parentId = entity.entityID.to_s
            parentId = entity.persistent_id.to_s
          end
          if root #最顶层的mergegroup结构
            flag = true
          else
            flag = (parentVisible and entity.visible? and entity.layer.visible?)
          end
          processEntities(entity.definition.entities,four_matrix_four(inputTrans,entity.transformation.to_a),curMat,parentId,false,flag,false)
        end
      end

      CHelper.updateBarInProcessEntities(processBar)
    end
  end
end